#include "Common.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "SExpr.hpp"
#include "Value.hpp"
#include "Compile.hpp"
#include "Eval.hpp"
#include "Cleanup.hpp"
#include "Protocol.hpp"

using std::string;
using std::vector;
using std::map;
using std::pair;

enum class CellType : uint32_t
{
    Empty,
    Num,
    Left,
    Right,
    Up,
    Down,
    Add,
    Sub,
    Mul,
    Div,
    Rem,
    Warp,
    Eq,
    Ne,
    S,
    A,
    B
};

class Cell
{
public:
    CellType cellType = CellType::Empty;
    mpz_class value;

    bool operator==(const Cell& other) const
    {
        if (cellType != other.cellType) return false;
        if (cellType == CellType::Num)
        {
            if (value != other.value) return false;
        }
        return true;
    }

    bool operator!=(const Cell& other) const
    {
        return !(*this == other);
    }
};

class State
{
public:
    bool init(const string& solution,
              const mpz_class* pA,
              const mpz_class* pB,
              string* pMsg = nullptr)
    {
        vector<string> lines;
        getLines(solution, &lines);

        height = lines.size();
        vector<vector<string>> tokens(height);
        for (int64_t y = 0; y < height; y++)
        {
            auto& line = lines[y];
            auto& lineTokens = tokens[y];
            bool inToken = false;
            for (char ch : line)
            {
                if (ch == ' ' || ch == '\t')
                {
                    inToken = false;
                }
                else
                {
                    if (!inToken)
                    {
                        inToken = true;
                        lineTokens.emplace_back();
                    }
                    lineTokens.back().push_back(ch);
                }
            }
            width = std::max(width, (int64_t)lineTokens.size());
        }

        length = 1;
        now = 0;
        maxNow = 0;
        cells.clear();
        cells.resize(1);
        cells[0].resize(width * height);

        int64_t minX = std::numeric_limits<int64_t>::max();
        int64_t maxX = std::numeric_limits<int64_t>::min();
        int64_t minY = std::numeric_limits<int64_t>::max();
        int64_t maxY = std::numeric_limits<int64_t>::min();

        for (int64_t y = 0; y < height; y++)
        {
            auto& lineTokens = tokens[y];
            for (int64_t x = 0; x < (int64_t)lineTokens.size(); x++)
            {
                auto& token = lineTokens[x];
                auto& cell = cells[0][y * width + x];
                int32_t literal = 0;
                if (token == ".") cell.cellType = CellType::Empty;
                else if (token == "<") cell.cellType = CellType::Left;
                else if (token == ">") cell.cellType = CellType::Right;
                else if (token == "^") cell.cellType = CellType::Up;
                else if (token == "v") cell.cellType = CellType::Down;
                else if (token == "+") cell.cellType = CellType::Add;
                else if (token == "-") cell.cellType = CellType::Sub;
                else if (token == "*") cell.cellType = CellType::Mul;
                else if (token == "/") cell.cellType = CellType::Div;
                else if (token == "%") cell.cellType = CellType::Rem;
                else if (token == "@") cell.cellType = CellType::Warp;
                else if (token == "=") cell.cellType = CellType::Eq;
                else if (token == "#") cell.cellType = CellType::Ne;
                else if (token == "S") cell.cellType = CellType::S;
                else if (token == "A")
                {
                    if (!pA)
                    {
                        if (pMsg) *pMsg = "No value specified for 'A'";
                        return false;
                    }
                    cell.cellType = CellType::Num;
                    cell.value = *pA;
                }
                else if (token == "B")
                {
                    if (!pB)
                    {
                        if (pMsg) *pMsg = "No value specified for 'B'";
                        return false;
                    }
                    cell.cellType = CellType::Num;
                    cell.value = *pB;
                }
                else if (parseI32(token, &literal) && literal >= -99 && literal <= 99)
                {
                    cell.cellType = CellType::Num;
                    cell.value = literal;
                }
                else
                {
                    if (pMsg) *pMsg = "Invalid token: '" + token + "'";
                    return false;
                }

                if (cell.cellType != CellType::Empty)
                {
                    minX = std::min(minX, x);
                    maxX = std::max(maxX, x);
                    minY = std::min(minY, y);
                    maxY = std::max(maxY, y);
                }
            }
        }

        if (minX > maxX || minY > maxY)
        {
            resize(0, 1, 0, 1);
        }
        else
        {
            resize(minX - 1, maxX + 1, minY - 1, maxY + 1);
        }

        done = false;
        result = Cell();
        score = 0;

        return true;
    }

    void print()
    {
        printf("time = %" PRIi64 "\n", now);
        for (int64_t y = 0; y < height; y++)
        {
            for (int64_t x = 0; x < width; x++)
            {
                auto& cell = cells[now][y * width + x];
                if (cell.cellType != CellType::Num) printf("   ");
                switch (cell.cellType)
                {
                case CellType::Empty: printf("."); break;
                case CellType::Num:
                    printf("%4s", cell.value.get_str().c_str());
                    break;
                case CellType::Left: printf("<"); break;
                case CellType::Right: printf(">"); break;
                case CellType::Up: printf("^"); break;
                case CellType::Down: printf("v"); break;
                case CellType::Add: printf("+"); break;
                case CellType::Sub: printf("-"); break;
                case CellType::Mul: printf("*"); break;
                case CellType::Div: printf("/"); break;
                case CellType::Rem: printf("%%"); break;
                case CellType::Warp: printf("@"); break;
                case CellType::Eq: printf("="); break;
                case CellType::Ne: printf("#"); break;
                case CellType::S: printf("S"); break;
                case CellType::A: printf("A"); break;
                case CellType::B: printf("B"); break;
                }
            }
            printf("\n");
        }
        printf("\n");
    }

    class Warp
    {
    public:
        int64_t newNow = 0;
        int64_t newX = 0;
        int64_t newY = 0;
        int64_t fromX = 0;
        int64_t fromY = 0;
    };

    bool step(string* pMsg = nullptr)
    {
        if (done)
        {
            if (pMsg) *pMsg = "Already done";
            return false;
        }

        vector<Warp> warps;

        int64_t newNow = now + 1;

        if (newNow == length)
        {
            cells.emplace_back();
            length++;
        }
        cells[newNow] = cells[now];

        vector<uint8_t> written(width * height);
        vector<uint8_t> toClear(width * height);

        bool didSomething = false;

        for (int64_t y = 0; y < height; y++)
        {
            for (int64_t x = 0; x < width; x++)
            {
                auto& cell = cells[now][y * width + x];
                if (cell.cellType == CellType::Empty)
                {
                    continue;
                }

                if (x > 0 && x + 1 < width)
                {
                    auto& cellL = cells[now][y * width + (x - 1)];
                    auto& cellR = cells[now][y * width + (x + 1)];
                    auto& newCellL = cells[newNow][y * width + (x - 1)];
                    auto& newCellR = cells[newNow][y * width + (x + 1)];
                    auto& writtenL = written[y * width + (x - 1)];
                    auto& writtenR = written[y * width + (x + 1)];
                    auto& toClearL = toClear[y * width + (x - 1)];
                    auto& toClearR = toClear[y * width + (x + 1)];

                    if (cell.cellType == CellType::Right &&
                        cellL.cellType != CellType::Empty)
                    {
                        if (writtenR)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenR = 1;
                        toClearL = 1;
                        newCellR = cellL;
                        didSomething = true;
                    }

                    if (cell.cellType == CellType::Left &&
                        cellR.cellType != CellType::Empty)
                    {
                        if (writtenL)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenL = 1;
                        toClearR = 1;
                        newCellL = cellR;
                        didSomething = true;
                    }
                }

                if (y > 0 && y + 1 < height)
                {
                    auto& cellU = cells[now][(y - 1) * width + x];
                    auto& cellD = cells[now][(y + 1) * width + x];
                    auto& newCellU = cells[newNow][(y - 1) * width + x];
                    auto& newCellD = cells[newNow][(y + 1) * width + x];
                    auto& writtenU = written[(y - 1) * width + x];
                    auto& writtenD = written[(y + 1) * width + x];
                    auto& toClearU = toClear[(y - 1) * width + x];
                    auto& toClearD = toClear[(y + 1) * width + x];

                    if (cell.cellType == CellType::Down &&
                        cellU.cellType != CellType::Empty)
                    {
                        if (writtenD)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenD = 1;
                        toClearU = 1;
                        newCellD = cellU;
                        didSomething = true;
                    }

                    if (cell.cellType == CellType::Up &&
                        cellD.cellType != CellType::Empty)
                    {
                        if (writtenU)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenU = 1;
                        toClearD = 1;
                        newCellU = cellD;
                        didSomething = true;
                    }
                }

                if (x > 0 && x + 1 < width &&
                    y > 0 && y + 1 < height)
                {
                    auto& cellL = cells[now][y * width + (x - 1)];
                    auto& cellR = cells[now][y * width + (x + 1)];
                    auto& cellU = cells[now][(y - 1) * width + x];
                    auto& cellD = cells[now][(y + 1) * width + x];
                    //auto& newCellL = cells[newNow][y * width + (x - 1)];
                    auto& newCellR = cells[newNow][y * width + (x + 1)];
                    //auto& newCellU = cells[newNow][(y - 1) * width + x];
                    auto& newCellD = cells[newNow][(y + 1) * width + x];
                    //auto& writtenL = written[y * width + (x - 1)];
                    auto& writtenR = written[y * width + (x + 1)];
                    //auto& writtenU = written[(y - 1) * width + x];
                    auto& writtenD = written[(y + 1) * width + x];
                    auto& toClearL = toClear[y * width + (x - 1)];
                    //auto& toClearR = toClear[y * width + (x + 1)];
                    auto& toClearU = toClear[(y - 1) * width + x];
                    //auto& toClearD = toClear[(y + 1) * width + x];

                    if ((cell.cellType == CellType::Add ||
                         cell.cellType == CellType::Sub ||
                         cell.cellType == CellType::Mul ||
                         cell.cellType == CellType::Div ||
                         cell.cellType == CellType::Rem) &&
                        cellL.cellType == CellType::Num &&
                        cellU.cellType == CellType::Num)
                    {
                        if (writtenR || writtenD)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenR = 1;
                        writtenD = 1;
                        toClearL = 1;
                        toClearU = 1;
                        newCellR.cellType = CellType::Num;
                        newCellD.cellType = CellType::Num;
                        if (cell.cellType == CellType::Add)
                        {
                            newCellR.value = cellL.value + cellU.value;
                        }
                        else if (cell.cellType == CellType::Sub)
                        {
                            newCellR.value = cellL.value - cellU.value;
                        }
                        else if (cell.cellType == CellType::Mul)
                        {
                            newCellR.value = cellL.value * cellU.value;
                        }
                        else if (cell.cellType == CellType::Div)
                        {
                            if (cellU.value == 0)
                            {
                                if (pMsg) *pMsg = "Division by zero";
                                return false;
                            }
                            newCellR.value = cellL.value / cellU.value;
                        }
                        else if (cell.cellType == CellType::Rem)
                        {
                            if (cellU.value == 0)
                            {
                                if (pMsg) *pMsg = "Division by zero";
                                return false;
                            }
                            newCellR.value = cellL.value % cellU.value;
                        }
                        newCellD.value = newCellR.value;
                        didSomething = true;
                    }

                    if (cellL.cellType != CellType::Empty &&
                        cellU.cellType != CellType::Empty &&
                        ((cell.cellType == CellType::Eq &&
                          cellL == cellU) ||
                         (cell.cellType == CellType::Ne &&
                          cellL != cellU)))
                    {
                        if (writtenR || writtenD)
                        {
                            if (pMsg) *pMsg = "Duplicate write";
                            return false;
                        }
                        writtenR = 1;
                        writtenD = 1;
                        toClearL = 1;
                        toClearU = 1;
                        newCellD = cellL;
                        newCellR = cellU;
                        didSomething = true;
                    }

                    if (cell.cellType == CellType::Warp &&
                        cellL.cellType == CellType::Num &&
                        cellR.cellType == CellType::Num &&
                        cellD.cellType == CellType::Num &&
                        cellU.cellType != CellType::Empty)
                    {
                        if (!cellL.value.fits_slong_p())
                        {
                            if (pMsg) *pMsg = "@: dx value out of range";
                            return false;
                        }
                        int64_t dx = cellL.value.get_si();

                        if (!cellR.value.fits_slong_p())
                        {
                            if (pMsg) *pMsg = "@: dy value out of range";
                            return false;
                        }
                        int64_t dy = cellR.value.get_si();

                        if (!cellD.value.fits_slong_p())
                        {
                            if (pMsg) *pMsg = "@: dt value out of range";
                            return false;
                        }
                        int64_t dt = cellD.value.get_si();
                        if (dt <= 0 || dt > now)
                        {
                            if (pMsg) *pMsg = "@: dt value out of range";
                            return false;
                        }

                        auto& warp = warps.emplace_back();
                        warp.newNow = now - dt;
                        warp.newX = x - dx;
                        warp.newY = y - dy;
                        warp.fromX = x;
                        warp.fromY = y - 1;
                        didSomething = true;
                    }
                }
            }
        }

        if (!didSomething)
        {
            if (pMsg) *pMsg = "Nothing to do";
            return false;
        }

        for (int64_t y = 0; y < height; y++)
        {
            for (int64_t x = 0; x < width; x++)
            {
                if (written[y * width + x] &&
                    cells[now][y * width + x].cellType == CellType::S)
                {
                    if (done)
                    {
                        if (pMsg) *pMsg = "Duplicate submit";
                        return false;
                    }
                    done = true;
                    result = cells[newNow][y * width + x];
                }
            }
        }

        if (!done && !warps.empty())
        {
            for (size_t i = 0; i < warps.size(); i++)
            {
                auto& warpI = warps[i];
                for (size_t j = 0; j < i; j++)
                {
                    auto& warpJ = warps[j];
                    if (warpJ.newNow != warpI.newNow)
                    {
                        if (pMsg) *pMsg = "Warping to different times";
                        return false;
                    }
                    if (warpJ.newX == warpI.newX &&
                        warpJ.newY == warpI.newY)
                    {
                        if (cells[now][warpI.fromY * width + warpI.fromX] !=
                            cells[now][warpJ.fromY * width + warpJ.fromX])
                        {
                            if (pMsg) *pMsg = "Warp writing different values to the same cell";
                            return false;
                        }
                    }
                }
            }
        }

        for (int64_t y = 0; y < height; y++)
        {
            for (int64_t x = 0; x < width; x++)
            {
                if (toClear[y * width + x] &&
                    !written[y * width + x])
                {
                    cells[newNow][y * width + x] = Cell();
                }
            }
        }

        if (!done && !warps.empty())
        {
            newNow = warps[0].newNow;
            int64_t minX = 1;
            int64_t maxX = width - 2;
            int64_t minY = 1;
            int64_t maxY = height - 2;
            for (size_t i = 0; i < warps.size(); i++)
            {
                auto& warpI = warps[i];
                if (warpI.newNow != newNow)
                {
                    if (pMsg) *pMsg = "Warping to different times";
                    return false;
                }
                minX = std::min(minX, warpI.newX);
                maxX = std::max(maxX, warpI.newX);
                minY = std::min(minY, warpI.newY);
                maxY = std::max(maxY, warpI.newY);
                for (size_t j = 0; j < i; j++)
                {
                    auto& warpJ = warps[j];
                    if (warpJ.newX == warpI.newX &&
                        warpJ.newY == warpI.newY)
                    {
                        if (cells[newNow][warpI.fromY * width + warpI.fromX] !=
                            cells[newNow][warpJ.fromY * width + warpJ.fromX])
                        {
                            if (pMsg) *pMsg = "Warp writing different values to the same cell";
                            return false;
                        }
                    }
                }
            }

            resize(minX - 1, maxX + 1, minY - 1, maxY + 1);

            for (auto& warp : warps)
            {
                warp.newX -= minX - 1;
                warp.newY -= minY - 1;
                cells[newNow][warp.newY * width + warp.newX] =
                    cells[now][warp.fromY * width + warp.fromX];
            }
        }

        now = newNow;

        if (!done && warps.empty())
        {
            int64_t minX = 1;
            int64_t maxX = width - 2;
            int64_t minY = 1;
            int64_t maxY = height - 2;

            for (int64_t y = 0; y < height; y++)
            {
                for (int64_t x = 0; x < width; x++)
                {
                    if (cells[now][y * width + x].cellType != CellType::Empty)
                    {
                        minX = std::min(minX, x);
                        maxX = std::max(maxX, x);
                        minY = std::min(minY, y);
                        maxY = std::max(maxY, y);
                    }
                }
            }

            resize(minX - 1, maxX + 1, minY - 1, maxY + 1);
        }

        if (!done)
        {
            maxNow = std::max(maxNow, now);
        }

        if (done)
        {
            score = (width - 2) * (height - 2) * (maxNow + 1);
        }

        return true;
    }

    void resize(int64_t minX, int64_t maxX, int64_t minY, int64_t maxY)
    {
        if (minX == 0 && maxX == width - 1 && minY == 0 && maxY == height - 1)
        {
            return;
        }

        int64_t newWidth = maxX - minX + 1;
        int64_t newHeight = maxY - minY + 1;

        for (int64_t t = 0; t < length; t++)
        {
            auto& oldGrid = cells[t];
            vector<Cell> newGrid(newWidth * newHeight);
            for (int64_t y = 0; y < height; y++)
            {
                int64_t newY = y - minY;
                if (newY < 0 || newY >= newHeight) continue;
                for (int64_t x = 0; x < width; x++)
                {
                    int64_t newX = x - minX;
                    if (newX < 0 || newX >= newWidth) continue;

                    auto& oldCell = oldGrid[y * width + x];
                    auto& newCell = newGrid[newY * newWidth + newX];
                    newCell = std::move(oldCell);
                }
            }
            oldGrid.swap(newGrid);
        }

        width = newWidth;
        height = newHeight;
    }

public:
    vector<vector<Cell>> cells;
    int64_t width = 0;
    int64_t height = 0;
    int64_t length = 0;
    int64_t now = 0;
    int64_t maxNow = 0;
    bool done = false;
    Cell result;
    int64_t score = 0;
};

enum class ArgType : uint32_t
{
    Var,
    Num,
    Empty,
    A,
    B,
    S
};

class Arg
{
public:
    Arg() {}
    Arg(ArgType argType_) : argType(argType_) {}
    Arg(const string& name_) : argType(ArgType::Var), name(name_) {}
    Arg(const char* name_) : argType(ArgType::Var), name(name_) {}
    Arg(int64_t value_) : argType(ArgType::Num), value(value_) {}
    Arg(int32_t value_) : argType(ArgType::Num), value(value_) {}

    ArgType argType = ArgType::Empty;
    string name;
    int64_t value = 0;
    int64_t varId = -1;
};

Arg E(ArgType::Empty);
Arg A(ArgType::A);
Arg B(ArgType::B);
Arg S(ArgType::S);

class Init
{
public:
    string name;
    int64_t varId = -1;
    Arg arg;
};

enum class InstType : uint32_t
{
    Invalid,
    Add,
    Sub,
    Mul,
    Div,
    Rem,
    Eq,
    Ne,
    Mov,
    Split2,
    Split3,
    Halt0,
    Halt1,
    HaltNeg1
};

class Inst
{
public:
    InstType instType = InstType::Invalid;
    vector<Arg> results;
    vector<Arg> args;
};

class Builder
{
public:
    void INIT(const string& name,
              const Arg& arg)
    {
        auto& init = inits.emplace_back();
        init.name = name;
        init.arg = arg;
    }

    void binOp(InstType instType,
               const Arg& result,
               const Arg& arg0,
               const Arg& arg1)
    {
        auto& inst = insts.emplace_back();
        inst.instType = instType;
        inst.results.push_back(result);
        inst.args.push_back(arg0);
        inst.args.push_back(arg1);
    }

    void ADD(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Add, result, arg0, arg1); }
    void SUB(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Sub, result, arg0, arg1); }
    void MUL(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Mul, result, arg0, arg1); }
    void DIV(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Div, result, arg0, arg1); }
    void REM(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Rem, result, arg0, arg1); }
    void EQ(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Eq, result, arg0, arg1); }
    void NE(const Arg& result, const Arg& arg0, const Arg& arg1)
    { binOp(InstType::Ne, result, arg0, arg1); }

    void MOV(const Arg& result,
             const Arg& arg)
    {
        auto& inst = insts.emplace_back();
        inst.instType = InstType::Mov;
        inst.results.push_back(result);
        inst.args.push_back(arg);
    }

    void SPLIT2(const Arg& result0,
                const Arg& result1,
                const Arg& arg)
    {
        auto& inst = insts.emplace_back();
        inst.instType = InstType::Split2;
        inst.results.push_back(result0);
        inst.results.push_back(result1);
        inst.args.push_back(arg);
    }

    void SPLIT3(const Arg& result0,
                const Arg& result1,
                const Arg& result2,
                const Arg& arg)
    {
        auto& inst = insts.emplace_back();
        inst.instType = InstType::Split3;
        inst.results.push_back(result0);
        inst.results.push_back(result1);
        inst.results.push_back(result2);
        inst.args.push_back(arg);
    }

    void haltOp(InstType instType,
                const Arg& arg0,
                const Arg& arg1)
    {
        auto& inst = insts.emplace_back();
        inst.instType = instType;
        inst.args.push_back(arg0);
        inst.args.push_back(arg1);
    }

    void HALT0(const Arg& arg0, const Arg& arg1)
    { haltOp(InstType::Halt0, arg0, arg1); }
    void HALT1(const Arg& arg0, const Arg& arg1)
    { haltOp(InstType::Halt1, arg0, arg1); }
    void HALTNEG1(const Arg& arg0, const Arg& arg1)
    { haltOp(InstType::HaltNeg1, arg0, arg1); }

    bool build(string* pData,
               string* pMsg = nullptr)
    {
        vector<string> idToName;
        map<string, int64_t> nameToId;
        int64_t varCount = 0;

        int64_t instCount = insts.size();
        for (int64_t instIdx = 0; instIdx < instCount; instIdx++)
        {
            auto& inst = insts[instIdx];
            for (int64_t resultIdx = 0; resultIdx < (int64_t)inst.results.size(); resultIdx++)
            {
                auto& result = inst.results[resultIdx];
                if (result.argType == ArgType::Var)
                {
                    int64_t varId = -1;
                    auto findIt = nameToId.find(result.name);
                    if (findIt != nameToId.end())
                    {
                        varId = findIt->second;
                    }
                    else
                    {
                        varId = varCount++;
                        idToName.push_back(result.name);
                        nameToId[result.name] = varId;
                    }
                    result.varId = varId;
                }
            }
        }

        vector<Arg> idToArg(varCount);

        int64_t initCount = inits.size();
        for (int64_t initIdx = 0; initIdx < initCount; initIdx++)
        {
            auto& init = inits[initIdx];
            auto findIt = nameToId.find(init.name);
            if (findIt == nameToId.end())
            {
                if (pMsg) *pMsg = "Initializing undefined variable: '" + init.name + "'";
                return false;
            }
            int64_t varId = findIt->second;
            init.varId = varId;
            idToArg[varId] = init.arg;
        }

        vector<uint8_t> varUsed(varCount);

        for (int64_t instIdx = 0; instIdx < instCount; instIdx++)
        {
            auto& inst = insts[instIdx];
            for (int64_t argIdx = 0; argIdx < (int64_t)inst.args.size(); argIdx++)
            {
                auto& arg = inst.args[argIdx];
                if (arg.argType == ArgType::Var)
                {
                    auto findIt = nameToId.find(arg.name);
                    if (findIt == nameToId.end())
                    {
                        if (pMsg) *pMsg = "Undefined variable '" + arg.name + "'";
                        return false;
                    }
                    int64_t argVarId = findIt->second;
                    arg.varId = argVarId;
                    if (varUsed[argVarId])
                    {
                        if (pMsg) *pMsg = "Variable used more than once: '" + arg.name + "'";
                        return false;
                    }
                    varUsed[argVarId] = 1;
                }
            }
        }

#if 0
        for (int64_t varId = 0; varId < varCount; varId++)
        {
            if (!varUsed[varId])
            {
                if (pMsg) *pMsg = "Unused variable: '" + idToName[varId] + "'";
                return false;
            }
        }
#endif

        vector<pair<int64_t, int64_t>> varUses(varCount, { -1, -1 });
        vector<vector<pair<int64_t, int64_t>>> varDefs(varCount);

        int64_t posX = 0;
        int64_t posY = 0;
        vector<vector<Cell>> cells(5);

        auto setEmpty = [&](int64_t offsetX, int64_t offsetY)
        {
            auto& cell = cells[posY + offsetY][posX + offsetX];
            cell = Cell();
        };
        auto setInt = [&](int64_t offsetX, int64_t offsetY, int64_t intValue)
        {
            auto& cell = cells[posY + offsetY][posX + offsetX];
            cell.cellType = CellType::Num;
            cell.value = (int32_t)intValue;
        };
        auto setArg = [&](int64_t offsetX, int64_t offsetY, Arg arg)
        {
            auto& cell = cells[posY + offsetY][posX + offsetX];

            if (arg.argType == ArgType::Var)
            {
                varUses[arg.varId] = { posX + offsetX, posY + offsetY };
                arg = idToArg[arg.varId];
            }

            switch (arg.argType)
            {
            case ArgType::Num:
                setInt(offsetX, offsetY, arg.value);
                break;
            case ArgType::Empty:
                setEmpty(offsetX, offsetY);
                break;
            case ArgType::A:
                cell.cellType = CellType::A;
                break;
            case ArgType::B:
                cell.cellType = CellType::B;
                break;
            default:
                cell.cellType = CellType::Left;
            }
        };
        auto setOp = [&](int64_t offsetX, int64_t offsetY, CellType cellType)
        {
            auto& cell = cells[posY + offsetY][posX + offsetX];
            cell.cellType = cellType;
        };
        auto setWarp = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::Warp);
        };
        auto setLeft = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::Left);
        };
        auto setRight = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::Right);
        };
#if 0
        auto setUp = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::Up);
        };
#endif
        auto setDown = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::Down);
        };
        auto setS = [&](int64_t offsetX, int64_t offsetY)
        {
            setOp(offsetX, offsetY, CellType::S);
        };

        for (int64_t instIdx = 0; instIdx < instCount; instIdx++)
        {
            auto& inst = insts[instIdx];

            int64_t instWidth = 0;

            switch (inst.instType)
            {
            case InstType::Add:
            case InstType::Sub:
            case InstType::Mul:
            case InstType::Div:
            case InstType::Rem:
            case InstType::Eq:
            case InstType::Ne:
            case InstType::Mov:
                instWidth = 3;
                break;
            case InstType::Split2:
                instWidth = 5;
                break;
            case InstType::Split3:
                instWidth = 7;
                break;
            case InstType::Halt0:
            case InstType::Halt1:
            case InstType::HaltNeg1:
                instWidth = 3;
                break;
            default:
                if (pMsg) *pMsg = "Instruction type not implemented";
                return false;
            }

            for (int64_t offsetY = 0; offsetY < 5; offsetY++)
            {
                for (int64_t offsetX = 0; offsetX < instWidth; offsetX++)
                {
                    cells[posY + offsetY].emplace_back();
                }
            }

            switch (inst.instType)
            {
            case InstType::Add:
            case InstType::Sub:
            case InstType::Mul:
            case InstType::Div:
            case InstType::Rem:
            case InstType::Eq:
            case InstType::Ne:
            {
                setArg(0, 1, inst.args[0]);
                setArg(1, 0, inst.args[1]);

                switch (inst.instType)
                {
                case InstType::Add: setOp(1, 1, CellType::Add); break;
                case InstType::Sub: setOp(1, 1, CellType::Sub); break;
                case InstType::Mul: setOp(1, 1, CellType::Mul); break;
                case InstType::Div: setOp(1, 1, CellType::Div); break;
                case InstType::Rem: setOp(1, 1, CellType::Rem); break;
                case InstType::Eq: setOp(1, 1, CellType::Eq); break;
                case InstType::Ne: setOp(1, 1, CellType::Ne); break;
                default: ;
                }

                auto& result = inst.results[0];
                if (result.argType == ArgType::Var)
                {
                    setWarp(1, 3);
                    varDefs[result.varId].emplace_back(posX + 1, posY + 3);
                    setInt(1, 4, 1);
                }
                else if (result.argType == ArgType::S)
                {
                    setS(1, 2);
                }

                break;
            }
            case InstType::Mov:
            {
                setArg(1, 0, inst.args[0]);
                setDown(1, 1);

                auto& result = inst.results[0];
                if (result.argType == ArgType::Var)
                {
                    setWarp(1, 3);
                    varDefs[result.varId].emplace_back(posX + 1, posY + 3);
                    setInt(1, 4, 1);
                }
                else if (result.argType == ArgType::S)
                {
                    setS(1, 2);
                }

                break;
            }
            case InstType::Split2:
            case InstType::Split3:
            {
                setArg(3, 0, inst.args[0]);
                setLeft(2, 0);
                setDown(3, 1);

                setWarp(1, 1);
                varDefs[inst.results[0].varId].emplace_back(posX + 1, posY + 1);
                setInt(1, 2, 1);

                setWarp(3, 3);
                varDefs[inst.results[1].varId].emplace_back(posX + 3, posY + 3);;
                setInt(3, 4, 1);

                if (inst.instType == InstType::Split3)
                {
                    setRight(4, 0);

                    setWarp(5, 1);
                    varDefs[inst.results[2].varId].emplace_back(posX + 5, posY + 1);
                    setInt(5, 2, 1);
                }
                break;
            }
            case InstType::Halt0:
            case InstType::Halt1:
            case InstType::HaltNeg1:
            {
                setArg(0, 1, inst.args[0]);
                switch (inst.instType)
                {
                case InstType::Halt0: setInt(1, 0, 0); break;
                case InstType::Halt1: setInt(1, 0, 1); break;
                case InstType::HaltNeg1: setInt(1, 0, -1); break;
                default: ;
                }
                setOp(1, 1, CellType::Eq);

                setArg(0, 3, inst.args[1]);
                switch (inst.instType)
                {
                case InstType::Halt0: setOp(1, 3, CellType::Add); break;
                case InstType::Halt1: setOp(1, 3, CellType::Mul); break;
                case InstType::HaltNeg1: setOp(1, 3, CellType::Mul); break;
                default: ;
                }
                setS(1, 4);

                break;
            }
            default:
                if (pMsg) *pMsg = "Instruction type not implemented";
                return false;
            }
#if 0
            if (inst.op == "=0" ||
                inst.op == "=1" ||
                inst.op == "=-1")
            {
                if (result.argType != ArgType::S)
                {
                    if (pMsg) *pMsg = "S result required for '=0' and '=1'";
                    return false;
                }

                setArg(0, 1, inst.args[0]);
                setInt(1, 0, inst.op == "=0" ? 0 : inst.op == "=1" ? 1 : -1);
                setOp(1, 1, "=");

                setArg(0, 3, inst.args[1]);
                setOp(1, 3, inst.op == "=0" ? "+" : "*");
                setS(1, 4);
            }
            else
            {
                setArg(1, 0, inst.args[1]);
                setArg(0, 1, inst.args[0]);
                setOp(1, 1, inst.op);

                if (result.argType == ArgType::Var)
                {
                    int64_t useCount = useCounts[result.varId];

                    if (useCount >= 1)
                    {
                        setWarp(1, 3);
                        setInt(1, 4, 1);
                    }

                    if (useCount == 1)
                    {       
                        varDefs[result.varId].emplace_back(posX + 1, posY + 3);
                    }
                    else if (useCount >= 2)
                    {
                        int64_t calcX = posX + 1;
                        int64_t calcY = posY + 3;
                        int64_t splitX = varSplits[result.varId].first;
                        int64_t splitY = varSplits[result.varId].second;
                        int64_t dx = calcX - splitX;
                        int64_t dy = calcY - splitY;
                        setInt(0, 3, dx);
                        setInt(2, 3, dy);
                    }
                }
                else if (result.argType == ArgType::S)
                {
                    setS(1, 2);
                }
            }
#endif

            posX += instWidth;
        }

        for (int64_t varId = 0; varId < varCount; varId++)
        {
            int64_t useX = varUses[varId].first;
            int64_t useY = varUses[varId].second;
            if (useX == -1 || useY == -1)
            {
                continue;
            }
            for (auto& def : varDefs[varId])
            {
                int64_t defX = def.first;
                int64_t defY = def.second;
                if (defX == -1 || defY == -1)
                {
                    continue;
                }
                int64_t dx = defX - useX;
                int64_t dy = defY - useY;
                cells[defY][defX - 1].cellType = CellType::Num;
                cells[defY][defX - 1].value = (int32_t)dx;
                cells[defY][defX + 1].cellType = CellType::Num;
                cells[defY][defX + 1].value = (int32_t)dy;
            }
        }

        string data;
        for (auto& row : cells)
        {
            for (auto& cell : row)
            {
                switch (cell.cellType)
                {
                case CellType::Empty: data += "   ."; break;
                case CellType::Num:
                    data += strprintf(" %3s", cell.value.get_str().c_str());
                    break;
                case CellType::Left: data += "   <"; break;
                case CellType::Right: data += "   >"; break;
                case CellType::Up: data += "   ^"; break;
                case CellType::Down: data += "   v"; break;
                case CellType::Add: data += "   +"; break;
                case CellType::Sub: data += "   -"; break;
                case CellType::Mul: data += "   *"; break;
                case CellType::Div: data += "   /"; break;
                case CellType::Rem: data += "   %"; break;
                case CellType::Warp: data += "   @"; break;
                case CellType::Eq: data += "   ="; break;
                case CellType::Ne: data += "   #"; break;
                case CellType::S: data += "   S"; break;
                case CellType::A: data += "   A"; break;
                case CellType::B: data += "   B"; break;
                default:
                    data += "   ?";
                };
            }
            data += '\n';
        }

        if (pData) *pData = std::move(data);
        return true;
    }

public:
    vector<Init> inits;
    vector<Inst> insts;
};

void usage()
{
    printf("Usage: 3d [<options>] <id> [<a> [<b>]]\n");
    printf("  <id>  Problem id\n");
    printf("  <a>   Input A\n");
    printf("  <b>   Input B\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -b    Build solution\n");
    printf("  -t    Test on server\n");
    printf("  -s    Submit to server\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotA = false;
    bool gotB = false;

    bool help = false;
    bool doBuild = false;
    bool doTest = false;
    bool doSubmit = false;
    uint64_t id = 0;
    int32_t iA = 0;
    int32_t iB = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-b")
        {
            doBuild = true;
        }
        else if (strArg == "-t")
        {
            doTest = true;
        }
        else if (strArg == "-s")
        {
            doSubmit = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id> 12)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotA)
        {
            if (!parseI32(strArg, &iA))
            {
                usage();
                return 1;
            }
            gotA = true;
        }
        else if (!gotB)
        {
            if (!parseI32(strArg, &iB))
            {
                usage();
                return 1;
            }
            gotB = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    if ((doBuild ? 1 : 0) + (doTest ? 1 : 0) + (doSubmit ? 1 : 0) > 1)
    {
        usage();
        return 1;
    }

    if (doBuild && (gotA || gotB))
    {
        usage();
        return 1;
    }

    if (doTest && (!gotA || !gotB))
    {
        usage();
        return 1;
    }

    if (doSubmit && (gotA || gotB))
    {
        usage();
        return 1;
    }

    mpz_class a;
    mpz_class b;
    if (gotA) a = iA;
    if (gotB) b = iB;

    mpz_class* pA = gotA ? &a : nullptr;
    mpz_class* pB = gotB ? &b : nullptr;

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string msg;

    if (doBuild)
    {
        Builder b;

        if (id == 1)
        {
#if 0
            b.INIT("n", A);
            b.INIT("prod", 1);
            b.INIT("done", E);

            b.INST("n", "-", "n", 1);
            b.INST("prod", "*", "prod", "n");
            b.INST("done", "=", "n", 1);
            b.INST(S, "*", "done", "prod");
#endif
#if 0
            b.INIT("n", A);
            b.INIT("prod", 1);

            b.INST("n", "-", "n", 1);
            b.INST("prod", "*", "prod", "n");
            b.INST(S, "=1", "n", "prod");
#endif
#if 1
            b.INIT("n", A);
            b.INIT("prod", 1);

            b.SPLIT3("n_1a", "n_1b", "n_1c", "n");
            b.SPLIT2("prod_1a", "prod_1b", "prod");

            b.SUB("n", "n_1a", 1);
            b.MUL("prod", "prod_1a", "n_1b");
            b.HALT1("n_1c", "prod_1b");
#endif
        }
        else if (id == 2)
        {
#if 0
            b.INIT("n", A);

            b.INST(S, "=", A, 0);

            b.INST("n", "/", "n", 2);

            b.INST(S, "=-1", "n", A);
            b.INST(S, "=1", "n", A);
#endif
        }
        else if (id == 3)
        {
#if 0
            b.INIT("n", A);

            b.INST(S, "=", A, 0);

            b.INST("n", "/", "n", 2);

            b.INST(S, "=", "n", -1);
            b.INST(S, "=", "n", 1);
#endif
        }
        else if (id == 4)
        {
            b.INIT("first", 1);
            b.INIT("n", 0);

            b.SUB("diff_2", A, B);
            b.DIV("n_2", "n", 2);
            b.SPLIT2("first_2a", "first_2b", "first");

            b.SUB("notfirst_3", 1, "first_2a");
            b.SPLIT2("first_3a", "first_3b", "first_2b");
            b.MOV("diff_3", "diff_2");
            b.MOV("n_3", "n_2");

            b.MUL("first_n_4", "first_3a", "diff_3");
            b.MUL("notfirst_n_4", "notfirst_3", "n_3");
            b.MUL("first_4", 0, "first_3b");

            b.ADD("n_5", "first_n_4", "notfirst_n_4");
            b.MOV("first_5", "first_4");

            b.SPLIT3("n_6a", "n_6b", "n", "n_5");
            b.MOV("first", "first_5");

            b.HALT1("n_6a", A);
            b.HALT0("n_6b", B);

#if 0
            b.INST("a", "-", A, 99);
            b.INST("b", "-", B, 99);

            b.INST("a", "+", "a", 1);
            b.INST("b", "+", "b", 1);

            b.INST(S, "=1", "a", A);
            b.INST(S, "=1", "b", B);
#endif
        }
        else if (id == 5)
        {
            b.INIT("n", A);

            b.SPLIT3("n_2a", "n_2b", "n_2c", "n");

            b.REM("ra", "n_2a", A);
            b.REM("rb", "n_2b", B);
            b.SPLIT2("n_3a", "n_3b", "n_2c");

            b.ADD("r", "ra", "rb");
            b.ADD("n", "n_3a", A);

            b.HALT0("r", "n_3b");
        }
        else if (id == 6)
        {
            b.INIT("n", A);

            b.SUB("n_2", "n", 1);

            b.SPLIT3("n_3a", "n_3b", "n", "n_2");

            b.HALT1("n_3a", 1);
            b.REM("r", A, "n_3b");

            b.HALT0("r", 0);
        }
        else if (id == 7)
        {
            b.INIT("left", A);
            b.INIT("right", 0);

            b.SPLIT3("left_1", "left_2", "left_3", "left");
            b.SPLIT3("right_1", "right_2", "right_3", "right");

            b.REM("digit", "left_1", 10);
            b.DIV("newLeft", "left_2", 10);
            b.MUL("shifted", "right_1", 10);
            b.SPLIT2("left_4", "left_5", "left_3");

            b.SPLIT2("newLeft_1", "left", "newLeft");
            b.ADD("right", "shifted", "digit");

            b.SUB("diff1", "left_4", "right_2");
            b.SUB("diff2", "newLeft_1", "right_3");

            b.HALT0("diff1", 1);
            b.HALT0("diff2", 1);
            b.HALT0("left_5", 0);

            b.REM("lastDigit", A, 10);
            b.HALT0("lastDigit", 0);
        }
#if 0
        else if (id == 9)
        {
            b.INIT("digits", A);
            b.INIT("count", 0);

            b.REM("digit", "digits", 10);
            b.DIV("rest", "digits", 10);

            b.ADD("more", "count", 1);
            b.SUB("less", "count", 1);

            b.SUB("check2", "digit", 2);

            b.ADD("sum", "digits", "count");
            b.HALT0("sum", 1);
        }
#endif
        else
        {
            printf("Problem %" PRIu64 " not implemented\n", id);
            return 1;
        }

        string buildData;
        if (!b.build(&buildData, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        printf("%s", buildData.c_str());

        return 0;
    }

    string solutionFileName = strprintf("../solutions/3d%" PRIu64 "", id);
    string solutionData;
    if (!readFile(solutionFileName, &solutionData))
    {
        printf("Error reading solution file\n");
        return 1;
    }

    if (doTest || doSubmit)
    {
        Value value;
        value.init();
        value->setValueType(ValueType::String);

        if (doTest)
        {
            value->stringData.stringValue =
                strprintf("test 3d %s %s\n",
                          a.get_str().c_str(),
                          b.get_str().c_str());
        }
        else
        {
            value->stringData.stringValue =
                strprintf("solve 3d%" PRIu64 "\n", id);
        }

        value->stringData.stringValue += solutionData;

        string requestStr;
        if (!encodeValue(value, &requestStr, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        string responseStr;
        if (!Protocol::communicate(requestStr, &responseStr, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Value responseValue;
        if (!decodeValue(responseStr, &responseValue, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Value evalValue;
        if (!eval(responseValue, &evalValue, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        if (evalValue->valueType == ValueType::String)
        {
            printf("%s\n", evalValue->stringData.stringValue.c_str());
        }
        else
        {
            string evalFormatted;
            if (!formatValue(evalValue, &evalFormatted, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }
            printf("%s\n", evalFormatted.c_str());
        }
    }
    else
    {
        State state;
        if (!state.init(solutionData, pA, pB, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        state.print();

        while (!state.done)
        {
            if (!state.step(&msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            state.print();
        }

        if (state.result.cellType != CellType::Num)
        {
            printf("result not a number\n");
        }
        else
        {
            printf("result = %s\n", state.result.value.get_str().c_str());
        }

        printf("score = %" PRIi64 "\n", state.score);
    }

    return 0;
}
