#include "Compile.hpp"
#include "SExpr.hpp"
#include "Value.hpp"

using std::string;
using std::vector;
using std::map;

namespace
{
    bool isInt(const string& str,
               mpz_class* pIntValue)
    {
        return pIntValue->set_str(str, 0) == 0;
    }

    bool isString(const string& str,
                  string* pStringValue)
    {
        size_t size = str.size();
        if (size < 2) return false;
        if (str.front() != '"' || str.back() != '"') return false;
        string stringValue;
        size_t pos = 1;
        while (pos < size - 1)
        {
            char ch = str[pos++];
            if (ch == '\\')
            {
                if (pos >= size - 1) return false;
                ch = str[pos++];
                if (ch == '\\')
                {
                    stringValue.push_back('\\');
                }
                else if (ch == 'n')
                {
                    stringValue.push_back('\n');
                }
                else
                {
                    return false;
                }
            }
            else
            {
                stringValue.push_back(ch);
            }
        }
        if (pStringValue) *pStringValue = stringValue;
        return true;
    }

    bool isUnaryOp(const string& str,
                   UnaryOp* pUnaryOp)
    {
        UnaryOp unaryOp = UnaryOp::Invalid;
        if (str == "-") unaryOp = UnaryOp::Negate;
        else if (str == "not") unaryOp = UnaryOp::Not;
        else if (str == "toInt") unaryOp = UnaryOp::ToInt;
        else if (str == "toString") unaryOp = UnaryOp::ToString;
        else return false;
        if (pUnaryOp) *pUnaryOp = unaryOp;
        return true;
    }

    bool isBinaryOp(const string& str,
                    BinaryOp* pBinaryOp)
    {
        BinaryOp binaryOp = BinaryOp::Invalid;
        if (str == "+") binaryOp = BinaryOp::Add;
        else if (str == "-") binaryOp = BinaryOp::Sub;
        else if (str == "*") binaryOp = BinaryOp::Mul;
        else if (str == "/") binaryOp = BinaryOp::Div;
        else if (str == "%") binaryOp = BinaryOp::Mod;
        else if (str == "<") binaryOp = BinaryOp::Lt;
        else if (str == ">") binaryOp = BinaryOp::Gt;
        else if (str == "=") binaryOp = BinaryOp::Eq;
        else if (str == "or") binaryOp = BinaryOp::Or;
        else if (str == "and") binaryOp = BinaryOp::And;
        else if (str == "concat") binaryOp = BinaryOp::Concat;
        else if (str == "take") binaryOp = BinaryOp::Take;
        else if (str == "drop") binaryOp = BinaryOp::Drop;
        else return false;
        if (pBinaryOp) *pBinaryOp = binaryOp;
        return true;
    }

    bool compileImpl(const SExpr& expr,
                     map<string, int64_t>& varNameToIdx,
                     int64_t varCount,
                     Value* pValue,
                     string* pMsg)
    {
        Value value;
        value.init();

        if (expr.m_type == SExprType::Symbol)
        {
            mpz_class intValue;
            string stringValue;
            if (expr.m_symbol == "true")
            {
                value->setValueType(ValueType::Bool);
                value->boolData.boolValue = true;
            }
            else if (expr.m_symbol == "false")
            {
                value->setValueType(ValueType::Bool);
                value->boolData.boolValue = false;
            }
            else if (isInt(expr.m_symbol, &intValue))
            {
                value->setValueType(ValueType::Int);
                value->intData.intValue = std::move(intValue);
            }
            else if (isString(expr.m_symbol, &stringValue))
            {
                value->setValueType(ValueType::String);
                value->stringData.stringValue = std::move(stringValue);
            }
            else
            {
                auto findIt = varNameToIdx.find(expr.m_symbol);
                if (findIt == varNameToIdx.end())
                {
                    if (pMsg) *pMsg = "Unbound variable '" + expr.m_symbol + '\'';
                    return false;
                }
                value->setValueType(ValueType::Var);
                value->varData.var = findIt->second;
            }
        }
        else if (expr.m_type == SExprType::List)
        {
            size_t size = expr.m_list.size();
            if (size == 0)
            {
                if (pMsg) *pMsg = "Unexpected empty list";
                return false;
            }

            bool bFirstSymbol = expr.m_list[0].m_type == SExprType::Symbol;
            string firstSymbol;
            if (bFirstSymbol) firstSymbol = expr.m_list[0].m_symbol;

            UnaryOp unaryOp = UnaryOp::Invalid;
            bool bUnary = bFirstSymbol && isUnaryOp(firstSymbol, &unaryOp);

            BinaryOp binaryOp = BinaryOp::Invalid;
            bool bBinary = bFirstSymbol && isBinaryOp(firstSymbol, &binaryOp);

            if (bUnary && bBinary)
            {
                if (size == 2) bBinary = false;
                else bUnary = false;
            }

            if (bUnary)
            {
                if (size != 2)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for unary operator";
                    return false;
                }
                value->setValueType(ValueType::Unary);
                value->unaryData.unaryOp = unaryOp;
                if (!compileImpl(expr.m_list[1], varNameToIdx, varCount, &value->unaryData.arg, pMsg))
                {
                    return false;
                }
            }
            else if (bBinary)
            {
#if 1
                if (size < 3)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for binary operator";
                    return false;
                }
                if (!compileImpl(expr.m_list[1], varNameToIdx, varCount, &value, pMsg)) return false;
                for (size_t idx = 2; idx < size; idx++)
                {
                    Value tmp;
                    tmp.init();
                    tmp->setValueType(ValueType::Binary);
                    tmp->binaryData.binaryOp = binaryOp;
                    tmp->binaryData.args[0] = std::move(value);
                    if (!compileImpl(expr.m_list[idx], varNameToIdx, varCount, &tmp->binaryData.args[1], pMsg)) return false;
                    value = std::move(tmp);
                }
#endif
#if 0
                if (size != 3)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for binary operator";
                    return false;
                }
                value->setValueType(ValueType::Binary);
                value->binaryData.binaryOp = binaryOp;
                if (!compileImpl(expr.m_list[1], varNameToIdx, varCount, &value->binaryData.args[0], pMsg) ||
                    !compileImpl(expr.m_list[2], varNameToIdx, varCount, &value->binaryData.args[1], pMsg))
                {
                    return false;
                }
#endif
            }
            else if (firstSymbol == "if")
            {
                if (size != 4)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for 'if'";
                    return false;
                }
                value->setValueType(ValueType::If);
                if (!compileImpl(expr.m_list[1], varNameToIdx, varCount, &value->ifData.condValue, pMsg) ||
                    !compileImpl(expr.m_list[2], varNameToIdx, varCount, &value->ifData.trueValue, pMsg) ||
                    !compileImpl(expr.m_list[3], varNameToIdx, varCount, &value->ifData.falseValue, pMsg))
                {
                    return false;
                }
            }
            else if (firstSymbol == "lambda")
            {
                if (size != 3)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for 'lambda'";
                    return false;
                }
                if (expr.m_list[1].m_type != SExprType::List)
                {
                    if (pMsg) *pMsg = "Lambda argument list is not list";
                    return false;
                }
                size_t paramCount = expr.m_list[1].m_list.size();
                vector<int64_t> oldVarIdxs(paramCount);
                Value* pBodyValue = &value;
                for (size_t paramIdx = 0; paramIdx < paramCount; paramIdx++)
                {
                    if (expr.m_list[1].m_list[paramIdx].m_type != SExprType::Symbol)
                    {
                        if (pMsg) *pMsg = "Lambda argument not symbol";
                        return false;
                    }
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_symbol;
                    auto findIt = varNameToIdx.find(paramName);
                    if (findIt == varNameToIdx.end())
                    {
                        oldVarIdxs[paramIdx] = -1;
                    }
                    else
                    {
                        oldVarIdxs[paramIdx] = findIt->second;
                    }
                    varNameToIdx[paramName] = varCount + paramIdx;
                    (*pBodyValue)->setValueType(ValueType::Lambda);
                    (*pBodyValue)->lambdaData.var = varCount + paramIdx;
                    (*pBodyValue)->lambdaData.body.init();
                    pBodyValue = &(*pBodyValue)->lambdaData.body;
                }
                if (!compileImpl(expr.m_list[2], varNameToIdx, varCount + paramCount, pBodyValue, pMsg)) return false;
                for (size_t paramIdx = paramCount; paramIdx > 0; )
                {
                    paramIdx--;
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_symbol;
                    if (oldVarIdxs[paramIdx] == -1)
                    {
                        varNameToIdx.erase(paramName);
                    }
                    else
                    {
                        varNameToIdx[paramName] = oldVarIdxs[paramIdx];
                    }
                }
            }
            else if (firstSymbol == "let")
            {
                if (size != 3)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for 'let'";
                    return false;
                }
                if (expr.m_list[1].m_type != SExprType::List)
                {
                    if (pMsg) *pMsg = "Let argument list is not list";
                    return false;
                }
                size_t paramCount = expr.m_list[1].m_list.size();
                vector<int64_t> oldVarIdxs(paramCount);
                Value* pBodyValue = &value;
                for (size_t paramIdx = 0; paramIdx < paramCount; paramIdx++)
                {
                    if (expr.m_list[1].m_list[paramIdx].m_type != SExprType::List)
                    {
                        if (pMsg) *pMsg = "Let argument not list";
                        return false;
                    }
                    if (expr.m_list[1].m_list[paramIdx].m_list.size() != 2)
                    {
                        if (pMsg) *pMsg = "Let argument has incorrect size";
                        return false;
                    }
                    if (expr.m_list[1].m_list[paramIdx].m_list[0].m_type != SExprType::Symbol)
                    {
                        if (pMsg) *pMsg = "Let argument name is not symbol";
                        return false;
                    }
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_list[0].m_symbol;
                    auto findIt = varNameToIdx.find(paramName);
                    if (findIt == varNameToIdx.end())
                    {
                        oldVarIdxs[paramIdx] = -1;
                    }
                    else
                    {
                        oldVarIdxs[paramIdx] = findIt->second;
                    }
                    varNameToIdx[paramName] = varCount + paramIdx;
                    (*pBodyValue)->setValueType(ValueType::Lambda);
                    (*pBodyValue)->lambdaData.var = varCount + paramIdx;
                    (*pBodyValue)->lambdaData.body.init();
                    pBodyValue = &(*pBodyValue)->lambdaData.body;
                }
                if (!compileImpl(expr.m_list[2], varNameToIdx, varCount + paramCount, pBodyValue, pMsg)) return false;
                for (size_t paramIdx = paramCount; paramIdx > 0; )
                {
                    paramIdx--;
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_list[0].m_symbol;
                    if (oldVarIdxs[paramIdx] == -1)
                    {
                        varNameToIdx.erase(paramName);
                    }
                    else
                    {
                        varNameToIdx[paramName] = oldVarIdxs[paramIdx];
                    }
                }
                for (size_t paramIdx = 0; paramIdx < paramCount; paramIdx++)
                {
                    Value applyValue;
                    applyValue.init();
                    applyValue->setValueType(ValueType::Binary);
                    applyValue->binaryData.binaryOp = BinaryOp::Apply;
                    applyValue->binaryData.args[0] = std::move(value);
                    if (!compileImpl(expr.m_list[1].m_list[paramIdx].m_list[1], varNameToIdx, varCount, &applyValue->binaryData.args[1], pMsg)) return false;
                    value = std::move(applyValue);
                }
            }
            else if (firstSymbol == "let*")
            {
                if (size != 3)
                {
                    if (pMsg) *pMsg = "Incorrect number of arguments for 'let'";
                    return false;
                }
                if (expr.m_list[1].m_type != SExprType::List)
                {
                    if (pMsg) *pMsg = "Let argument list is not list";
                    return false;
                }
                size_t paramCount = expr.m_list[1].m_list.size();
                vector<int64_t> oldVarIdxs(paramCount);
                Value* pBodyValue = &value;
                for (size_t paramIdx = 0; paramIdx < paramCount; paramIdx++)
                {
                    if (expr.m_list[1].m_list[paramIdx].m_type != SExprType::List)
                    {
                        if (pMsg) *pMsg = "Let argument not list";
                        return false;
                    }
                    if (expr.m_list[1].m_list[paramIdx].m_list.size() != 2)
                    {
                        if (pMsg) *pMsg = "Let argument has incorrect size";
                        return false;
                    }
                    if (expr.m_list[1].m_list[paramIdx].m_list[0].m_type != SExprType::Symbol)
                    {
                        if (pMsg) *pMsg = "Let argument name is not symbol";
                        return false;
                    }
                    {
                        (*pBodyValue)->setValueType(ValueType::Binary);
                        (*pBodyValue)->binaryData.binaryOp = BinaryOp::Apply;
                        if (!compileImpl(expr.m_list[1].m_list[paramIdx].m_list[1], varNameToIdx, varCount, &(*pBodyValue)->binaryData.args[1], pMsg)) return false;
                        (*pBodyValue)->binaryData.args[0].init();
                        pBodyValue = &(*pBodyValue)->binaryData.args[0];
                    }
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_list[0].m_symbol;
                    auto findIt = varNameToIdx.find(paramName);
                    if (findIt == varNameToIdx.end())
                    {
                        oldVarIdxs[paramIdx] = -1;
                    }
                    else
                    {
                        oldVarIdxs[paramIdx] = findIt->second;
                    }
                    varNameToIdx[paramName] = varCount;
                    (*pBodyValue)->setValueType(ValueType::Lambda);
                    (*pBodyValue)->lambdaData.var = varCount;
                    (*pBodyValue)->lambdaData.body.init();
                    pBodyValue = &(*pBodyValue)->lambdaData.body;
                    varCount++;
                }
                if (!compileImpl(expr.m_list[2], varNameToIdx, varCount, pBodyValue, pMsg)) return false;
                for (size_t paramIdx = paramCount; paramIdx > 0; )
                {
                    paramIdx--;
                    const string& paramName = expr.m_list[1].m_list[paramIdx].m_list[0].m_symbol;
                    if (oldVarIdxs[paramIdx] == -1)
                    {
                        varNameToIdx.erase(paramName);
                    }
                    else
                    {
                        varNameToIdx[paramName] = oldVarIdxs[paramIdx];
                    }
                }
                varCount -= paramCount;
            }
            else
            {
                if (!compileImpl(expr.m_list[0], varNameToIdx, varCount, &value, pMsg)) return false;
                for (size_t i = 1; i < size; i++)
                {
                    Value applyValue;
                    applyValue.init();
                    applyValue->setValueType(ValueType::Binary);
                    applyValue->binaryData.binaryOp = BinaryOp::Apply;
                    applyValue->binaryData.args[0] = std::move(value);
                    if (!compileImpl(expr.m_list[i], varNameToIdx, varCount, &applyValue->binaryData.args[1], pMsg)) return false;
                    value = std::move(applyValue);
                }
            }
        }
        else
        {
            if (pMsg) *pMsg = "Unexpected expression type";
            return false;
        }

        if (pValue) *pValue = std::move(value);
        return true;
    }
}

bool compile(const SExpr& expr,
             Value* pValue,
             string* pMsg)
{
    map<string, int64_t> varNameToIdx;
    Value value;
    if (!compileImpl(expr, varNameToIdx, 0, &value, pMsg)) return false;
    if (pValue) *pValue = std::move(value);
    return true;
}
