#include "DebugGraphics.hpp"
#include "Graphics.hpp"
#include "DrawProblem.hpp"
#include "TimeUtils.hpp"

using std::string;
using std::vector;

DebugGraphics::DebugGraphics()
{
}

DebugGraphics::~DebugGraphics()
{
}

void DebugGraphics::create()
{
    auto threadFunc = [this]()
    {
        bool destroyRequested = false;
        bool destroyed = false;
        Graphics gr;

        Problem problem;
        vector<Coord> fakeSolution;
        string text;

        auto onDestroy = [&]()
        {
            destroyed = true;
            {
                LockGuard<Mutex> lock(m_mutex);
                m_destroyed = destroyed;
            }
        };
        gr.setOnDestroy(onDestroy);

        auto onKey = [&](Key key)
        {
            switch (key)
            {
            case Key::Esc:
            case Key::Q:
            case Key::ShiftQ:
            {
                if (!destroyRequested)
                {
                    destroyRequested = true;
                    gr.destroy();
                }
                break;
            }
            default: ;
            }
        };
        gr.setOnKey(onKey);

        auto onPaint = [&]()
        {
            gr.beginPaint(true);

            drawProblem(gr, problem, fakeSolution, text);

            gr.endPaint();
        };
        gr.setOnPaint(onPaint);

        string msg;
        if (!gr.create(1000, 1000, "Debug Graphics", &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            exit(1);
        }

        while (!destroyed)
        {
            if (Graphics::update(false))
            {
                continue;
            }

            sleepMS(10);

            {
                LockGuard<Mutex> lock(m_mutex);

                if (m_destroy)
                {
                    if (!destroyRequested)
                    {
                        destroyRequested = true;
                        gr.destroy();
                    }
                }
                else if (m_update)
                {
                    m_update = false;
                    if (m_updateProblem)
                    {
                        m_updateProblem = false;
                        problem = m_problem;
                    }
                    fakeSolution = m_fakeSolution;
                    text = m_text;
                    gr.requestPaint();
                }
            }
        }
    };

    m_thread = Thread(threadFunc);
}

void DebugGraphics::destroy()
{
    {
        LockGuard<Mutex> lock(m_mutex);
        m_destroy = true;
    }

    if (m_thread.joinable())
    {
        m_thread.join();
    }
}

bool DebugGraphics::destroyed()
{
    bool destroyed;
    {
        LockGuard<Mutex> lock(m_mutex);
        destroyed = m_destroyed;
    }
    return destroyed;
}

void DebugGraphics::update(const Problem& problem,
                           const vector<Coord>& fakeSolution,
                           const string& text)
{
    LockGuard<Mutex> lock(m_mutex);
    m_problem = problem;
    m_fakeSolution = fakeSolution;
    m_text = text;
    m_update = true;
    m_updateProblem = true;
}

void DebugGraphics::update(const vector<Coord>& fakeSolution,
                           const string& text)
{
    LockGuard<Mutex> lock(m_mutex);
    m_fakeSolution = fakeSolution;
    m_text = text;
    m_update = true;
}
