#include "DrawProblem.hpp"
#include "Graphics.hpp"
#include "Problem.hpp"
#include "Sim.hpp"
#include <limits>

using std::string;
using std::vector;

void drawProblemImpl(Graphics& gr,
                     const Problem& problem,
                     const Solution* pSolution,
                     const vector<Coord>* pFakeSolution,
                     const string& text)
{
    uint32_t width = gr.getWidth();
    uint32_t height = gr.getHeight();

    gr.setFont(FontSpacing::Monospace, FontFace::SansSerif,
               18, false, false);

    gr.fillRect(0, 0, width, height, 0x404040);

    if (!text.empty())
    {
        gr.drawText(2, 0,
                    text.c_str(),
                    0xffffff,
                    HAlign::Left, VAlign::Top);
    }

    uint32_t margin = 40;
    if (width <= margin * 2 || height <= margin * 2) return;

    int64_t minX = 0;
    int64_t maxX = 0;
    int64_t minY = 0;
    int64_t maxY = 0;

    for (Coord coord : problem)
    {
        minX = std::min(minX, coord.x);
        maxX = std::max(maxX, coord.x);
        minY = std::min(minY, coord.y);
        maxY = std::max(maxY, coord.y);
    }

#if 0
    minX -= 5;
    maxX += 5;
    minY -= 5;
    maxX += 5;
#endif

    int64_t worldMidX = (minX + maxX) / 2;
    int64_t worldMidY = (minY + maxY) / 2;
    int64_t screenMidX = (int64_t)width / 2;
    int64_t screenMidY = (int64_t)height / 2;

    int64_t problemWidth = maxX - minX;
    int64_t problemHeight = maxY - minY;
    int64_t problemSize = std::max(problemWidth, problemHeight);
    int64_t screenSize = std::min((int64_t)(width - margin * 2), (int64_t)(height - margin * 2));

    auto convertDist = [&](int64_t worldDist) -> int32_t
    {
        return (int32_t)(worldDist * screenSize / problemSize);
    };
    auto convertX = [&](int64_t worldX) -> int32_t
    {
        return (int32_t)((worldX - worldMidX) * screenSize / problemSize + screenMidX);
    };
    auto convertY = [&](int64_t worldY) -> int32_t
    {
        return (int32_t)(-(worldY - worldMidY) * screenSize / problemSize + screenMidY);
    };

    gr.fillRect(convertX(minX),
                convertY(maxY),
                convertDist(maxX - minX),
                convertDist(maxY - minY),
                0x000000);

    for (Coord coord : problem)
    {
        gr.fillCircle(convertX(coord.x) - 5,
                      convertY(coord.y) - 5,
                      10,
                      10,
                      0xff0000);
    }

    gr.fillCircle(convertX(0) - 4,
                  convertY(0) - 4,
                  8,
                  8,
                  0x0000ff);

    if (pSolution)
    {
        Ship ship;
        for (char move : *pSolution)
        {
            Ship newShip(ship);
            newShip.step(move);

            gr.drawLine(convertX(ship.x),
                        convertY(ship.y),
                        convertX(newShip.x),
                        convertY(newShip.y),
                        0xffffff);

            ship = newShip;
        }
    }

    if (pFakeSolution)
    {
        for (uint64_t idx = 1; idx < pFakeSolution->size(); idx++)
        {
            Coord oldCoord = (*pFakeSolution)[idx - 1];
            Coord newCoord = (*pFakeSolution)[idx];
            gr.drawLine(convertX(oldCoord.x),
                        convertY(oldCoord.y),
                        convertX(newCoord.x),
                        convertY(newCoord.y),
                        0xffffff);
        }
    }
}

void drawProblem(Graphics& gr,
                 const Problem& problem,
                 const Solution& solution,
                 const string& text)
{
    drawProblemImpl(gr, problem, &solution, nullptr, text);
}

void drawProblem(Graphics& gr,
                 const Problem& problem,
                 const vector<Coord>& fakeSolution,
                 const string& text)
{
    drawProblemImpl(gr, problem, nullptr, &fakeSolution, text);
}
