#include "Eval.hpp"
#include "Value.hpp"
#include "Cleanup.hpp"
#include <limits>
#include <sys/time.h>
#include <sys/resource.h>

using std::string;
using std::vector;
using std::map;
using std::unique_ptr;

#define USE_TAIL_CALL 0

namespace
{
    class StackSizeUpdater
    {
    public:
        StackSizeUpdater()
        {
            struct rlimit rlim;
            if (getrlimit(RLIMIT_STACK, &rlim) != 0)
            {
                fprintf(stderr, "getrlimit failed\n");
                exit(1);
            }
            rlim_t desiredStack = 1024 * 1024 * 1024;
            if (rlim.rlim_cur < desiredStack) rlim.rlim_cur = desiredStack;
            if (setrlimit(RLIMIT_STACK, &rlim) != 0)
            {
                fprintf(stderr, "setrlimit failed\n");
                exit(1);
            }
        }
    } stackSizeUpdater;

    bool eval(Value value,
              vector<Value>& varsIn,
              Value* pResult,
              string* pMsg)
    {
#if 0
        static uint64_t depth = 0;
        depth++;
        Cleanup cleanupDepth([&](){ depth--; });
        printf("depth = %" PRIu64 "\n", depth);
#endif

        Value result;
        result.init();

        vector<Value> localVars;
        vector<Value> newVars;
        vector<Value>* pVars = &varsIn;

        while (true)
        {
            auto& vars = *pVars;

            if (!value)
            {
                if (pMsg) *pMsg = "Null value";
                return false;
            }

            switch (value->valueType)
            {
            case ValueType::Invalid:
            {
                if (pMsg) *pMsg = "Invalid value";
                return false;
            }
            case ValueType::Bool:
            case ValueType::Int:
            case ValueType::String:
            {
                //printf("EVAL: bool/int/string\n");
                result = value;
                break;
            }
            case ValueType::Unary:
            {
                //printf("EVAL: unary %" PRIu32 "\n", (uint32_t)value->unaryData.unaryOp);
                switch (value->unaryData.unaryOp)
                {
                case UnaryOp::Invalid:
                {
                    if (pMsg) *pMsg = "Invalid unary operator";
                    return false;
                }
                case UnaryOp::Negate:
                {
                    Value argValue;
                    if (!eval(value->unaryData.arg, vars, &argValue, pMsg)) return false;
                    if (argValue->valueType != ValueType::Int)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const mpz_class& a = argValue->intData.intValue;
                    result->setValueType(ValueType::Int);
                    mpz_class& r = result->intData.intValue;
                    r = -a;
                    break;
                }
                case UnaryOp::Not:
                {
                    Value argValue;
                    if (!eval(value->unaryData.arg, vars, &argValue, pMsg)) return false;
                    if (argValue->valueType != ValueType::Bool)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    bool a = argValue->boolData.boolValue;
                    bool r = !a;
                    result->setValueType(ValueType::Bool);
                    result->boolData.boolValue = r;
                    break;
                }
                case UnaryOp::ToInt:
                {
                    Value argValue;
                    if (!eval(value->unaryData.arg, vars, &argValue, pMsg)) return false;
                    if (argValue->valueType != ValueType::String)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const string& a = argValue->stringData.stringValue;
                    result->setValueType(ValueType::Int);
                    mpz_class& r = result->intData.intValue;
                    string encodedStr;
                    if (!encodeString(a, encodedStr, pMsg)) return false;
                    if (!decodeInt(encodedStr, 0, &r, pMsg)) return false;
                    break;
                }
                case UnaryOp::ToString:
                {
                    Value argValue;
                    if (!eval(value->unaryData.arg, vars, &argValue, pMsg)) return false;
                    if (argValue->valueType != ValueType::Int)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const mpz_class& a = argValue->intData.intValue;
                    result->setValueType(ValueType::String);
                    string& r = result->stringData.stringValue;
                    string encodedStr;
                    if (!encodeInt(a, encodedStr, pMsg)) return false;
                    if (!decodeString(encodedStr, 0, &r, pMsg)) return false;
                    break;
                }
                }
                break;
            }
            case ValueType::Binary:
            {
                //printf("EVAL: binary %" PRIu32 "\n", (uint32_t)value->binaryData.binaryOp);
                switch (value->binaryData.binaryOp)
                {
                case BinaryOp::Invalid:
                {
                    if (pMsg) *pMsg = "Invalid binary operator";
                    return false;
                }
                case BinaryOp::Add:
                case BinaryOp::Sub:
                case BinaryOp::Mul:
                case BinaryOp::Div:
                case BinaryOp::Mod:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                    if (argValues[0]->valueType != ValueType::Int ||
                        argValues[1]->valueType != ValueType::Int)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const mpz_class& a = argValues[0]->intData.intValue;
                    const mpz_class& b = argValues[1]->intData.intValue;
                    result->setValueType(ValueType::Int);
                    mpz_class& r = result->intData.intValue;
                    switch (value->binaryData.binaryOp)
                    {
                    case BinaryOp::Add: r = a + b; break;
                    case BinaryOp::Sub: r = a - b; break;
                    case BinaryOp::Mul: r = a * b; break;
                    case BinaryOp::Div:
                    {
                        if (b == 0)
                        {
                            if (pMsg) *pMsg = "Division by zero";
                            return false;
                        }
                        r = a / b;
                        break;
                    }
                    case BinaryOp::Mod:
                    {
                        if (b == 0)
                        {
                            if (pMsg) *pMsg = "Modulo by zero";
                            return false;
                        }
                        r = a % b;
                        break;
                    }
                    default: ;
                    }
                    break;
                }
                case BinaryOp::Lt:
                case BinaryOp::Gt:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                    if (argValues[0]->valueType != ValueType::Int ||
                        argValues[1]->valueType != ValueType::Int)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const mpz_class& a = argValues[0]->intData.intValue;
                    const mpz_class& b = argValues[1]->intData.intValue;
                    bool r = false;
                    switch (value->binaryData.binaryOp)
                    {
                    case BinaryOp::Lt: r = a < b; break;
                    case BinaryOp::Gt: r = a > b; break;
                    default: ;
                    }
                    result->setValueType(ValueType::Bool);
                    result->boolData.boolValue = r;
                    break;
                }
                case BinaryOp::Eq:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                    bool r;
                    if (argValues[0]->valueType == ValueType::Int &&
                        argValues[1]->valueType == ValueType::Int)
                    {
                        const mpz_class& a = argValues[0]->intData.intValue;
                        const mpz_class& b = argValues[1]->intData.intValue;
                        r = a == b;
                    }
                    else if (argValues[0]->valueType == ValueType::Bool &&
                             argValues[1]->valueType == ValueType::Bool)
                    {
                        bool a = argValues[0]->boolData.boolValue;
                        bool b = argValues[1]->boolData.boolValue;
                        r = a == b;
                    }
                    else if (argValues[0]->valueType == ValueType::String &&
                             argValues[1]->valueType == ValueType::String)
                    {
                        const string& a = argValues[0]->stringData.stringValue;
                        const string& b = argValues[1]->stringData.stringValue;
                        r = a == b;
                    }
                    else
                    {
                        if (pMsg) *pMsg = "Bad argument types";
                        return false;
                    }
                    result->setValueType(ValueType::Bool);
                    result->boolData.boolValue = r;
                    break;
                }
                case BinaryOp::Or:
                case BinaryOp::And:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (argValues[0]->valueType != ValueType::Bool)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    bool a = argValues[0]->boolData.boolValue;
                    bool r = a;
                    bool done = false;
                    switch (value->binaryData.binaryOp)
                    {
                    case BinaryOp::Or: if (a) { done = true; } break;
                    case BinaryOp::And: if (!a) { done = true; } break;
                    default: ;
                    }
                    if (!done)
                    {
                        if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                        if (argValues[1]->valueType != ValueType::Bool)
                        {
                            if (pMsg) *pMsg = "Bad argument type";
                            return false;
                        }
                        bool b = argValues[1]->boolData.boolValue;
                        r = b;
                    }
                    result->setValueType(ValueType::Bool);
                    result->boolData.boolValue = r;
                    break;
                }
                case BinaryOp::Concat:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                    if (argValues[0]->valueType != ValueType::String ||
                        argValues[1]->valueType != ValueType::String)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const string& a = argValues[0]->stringData.stringValue;
                    const string& b = argValues[1]->stringData.stringValue;
                    result->setValueType(ValueType::String);
                    string& r = result->stringData.stringValue;
                    r = a + b;
                    break;
                }
                case BinaryOp::Take:
                case BinaryOp::Drop:
                {
                    Value argValues[2];
                    if (!eval(value->binaryData.args[0], vars, &argValues[0], pMsg)) return false;
                    if (!eval(value->binaryData.args[1], vars, &argValues[1], pMsg)) return false;
                    if (argValues[0]->valueType != ValueType::Int ||
                        argValues[1]->valueType != ValueType::String)
                    {
                        if (pMsg) *pMsg = "Bad argument type";
                        return false;
                    }
                    const mpz_class& a = argValues[0]->intData.intValue;
                    const string& b = argValues[1]->stringData.stringValue;
                    result->setValueType(ValueType::String);
                    string& r = result->stringData.stringValue;
                    switch (value->binaryData.binaryOp)
                    {
                    case BinaryOp::Take:
                    {
                        if (a < 0)
                        {
                            if (pMsg) *pMsg = "Take argument negative";
                            return false;
                        }
                        if (!a.fits_ulong_p())
                        {
                            if (pMsg) *pMsg = "Take argument too large";
                            return false;
                        }
                        unsigned long ulA = a.get_ui();
                        if (ulA > b.size())
                        {
                            if (pMsg) *pMsg = "Take argument too large";
                            return false;
                        }
                        r = b.substr(0, ulA);
                        break;
                    }
                    case BinaryOp::Drop:
                    {
                        if (a < 0)
                        {
                            if (pMsg) *pMsg = "Drop argument negative";
                            return false;
                        }
                        if (!a.fits_ulong_p())
                        {
                            if (pMsg) *pMsg = "Take argument too large";
                            return false;
                        }
                        unsigned long ulA = a.get_ui();
                        if (ulA > b.size())
                        {
                            if (pMsg) *pMsg = "Drop argument too large";
                            return false;
                        }
                        r = b.substr(ulA);
                        break;
                    }
                    default: ;
                    }
                    break;
                }
                case BinaryOp::Apply:
                {
                    Value funcValue;
                    if (!eval(value->binaryData.args[0], vars, &funcValue, pMsg)) return false;
                    if (funcValue->valueType != ValueType::Closure)
                    {
                        if (pMsg) *pMsg = "Apply argument not closure";
                        return false;
                    }

                    newVars = funcValue->closureData.vars;
                    auto& thunk = newVars.emplace_back();
                    thunk.init();
                    thunk->setValueType(ValueType::Thunk);
                    thunk->thunkData.vars = vars;
                    thunk->thunkData.body = value->binaryData.args[1];

#if USE_TAIL_CALL
                    value = funcValue->closureData.body;
                    localVars.swap(newVars);
                    pVars = &localVars;
                    continue;
#else
                    if (!eval(funcValue->closureData.body, newVars, &result, pMsg)) return false;
#endif
                }
                }
                break;
            }
            case ValueType::If:
            {
                //printf("EVAL: if\n");
                Value condValue;
                if (!eval(value->ifData.condValue, vars, &condValue, pMsg)) return false;
                if (condValue->valueType != ValueType::Bool)
                {
                    if (pMsg) *pMsg = "If conditional not boolean";
                    return false;
                }
                bool cond = condValue->boolData.boolValue;
                if (cond)
                {
#if USE_TAIL_CALL
                    value = value->ifData.trueValue;
                    continue;
#else
                    if (!eval(value->ifData.trueValue, vars, &result, pMsg)) return false;
#endif
                }
                else
                {
#if USE_TAIL_CALL
                    value = value->ifData.falseValue;
                    continue;
#else
                    if (!eval(value->ifData.falseValue, vars, &result, pMsg)) return false;
#endif
                }
                break;
            }
            case ValueType::Lambda:
            {
                //printf("EVAL: lambda\n");
                result->setValueType(ValueType::Closure);
                result->closureData.vars = vars;
                result->closureData.body = value->lambdaData.body;
                break;
            }
            case ValueType::Var:
            {
                //printf("EVAL: var\n");
                int64_t varIdx = value->varData.var;
                if (varIdx < 0 || varIdx >= (int64_t)vars.size())
                {
                    if (pMsg) *pMsg = "Variable index out of range";
                    return false;
                }
                Value& varValue = vars[varIdx];
                if (!eval(varValue, vars, &result, pMsg)) return false;
#if 1
                varValue = result;
#endif
                break;
            }
            case ValueType::Closure:
            {
                //printf("EVAL: closure\n");
                result = value;
                break;
            }
            case ValueType::Thunk:
            {
                //printf("EVAL: thunk\n");
#if USE_TAIL_CALL
                localVars = value->thunkData.vars;
                value = value->thunkData.body;
                pVars = &localVars;
                continue;
#else
                if (!eval(value->thunkData.body, value->thunkData.vars, &result, pMsg)) return false;
#endif
                break;
            }
            default:
            {
                if (pMsg) *pMsg = "Unsupported value type";
                return false;
            }
            }

            if (!result)
            {
                if (pMsg) *pMsg = "About to return null result";
                return false;
            }

            if (pResult) *pResult = std::move(result);
            return true;
        }
    }
}

bool eval(const Value& value,
          Value* pResult,
          string* pMsg)
{
    vector<Value> vars;
    Value result;
    if (!eval(value, vars, &result, pMsg)) return false;
    if (pResult) *pResult = std::move(result);
    return true;
}
