#include "SExpr.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

SExpr::SExpr() :
    m_type(SExprType::List),
    m_list(),
    m_symbol()
{
}

SExpr::SExpr(SExprType type) :
    m_type(type),
    m_list(),
    m_symbol()
{
}

SExpr::SExpr(string symbol) :
    m_type(SExprType::Symbol),
    m_list(),
    m_symbol(std::move(symbol))
{
}

SExpr::~SExpr()
{
}

namespace
{
    void formatIndent(string& str, uint32_t indent)
    {
        for (uint32_t i = 0; i < indent; i++)
        {
            str += ' ';
        }
    }

    void formatSExprImpl(const SExpr& expr,
                         string& str,
                         uint32_t indent)
    {
        if (expr.m_type == SExprType::List)
        {
            bool allSimple = true;
            for (size_t i = 0; i < expr.m_list.size(); i++)
            {
                if (expr.m_list[i].m_type != SExprType::Symbol)
                {
                    allSimple = false;
                    break;
                }
            }

            if (allSimple)
            {
                formatIndent(str, indent);
                str += '(';

                for (size_t i = 0; i < expr.m_list.size(); i++)
                {
                    if (i > 0)
                    {
                        str += ' ';
                    }
                    str += expr.m_list[i].m_symbol;;
                }

                str += ')';
                str += '\n';
            }
            else
            {
                formatIndent(str, indent);
                str += '(';
                str += '\n';

                for (size_t i = 0; i < expr.m_list.size(); i++)
                {
                    formatSExprImpl(expr.m_list[i], str, indent + 3);
                }

                formatIndent(str, indent);
                str += ')';
                str += '\n';
            }
        }
        else if (expr.m_type == SExprType::Symbol)
        {
            formatIndent(str, indent);
            str += expr.m_symbol;
            str += '\n';
        }
    }
}

void formatSExpr(const SExpr& expr,
                 string* pStr)
{
    string str;
    formatSExprImpl(expr, str, 0);
    if (pStr) *pStr = str;
}

namespace
{
    enum class State : uint32_t
    {
        Before,
        During,
        After,
        InQuote
    };
}

bool parseSExpr(const string& str,
                SExpr* pExpr,
                string* pMsg)
{
    size_t pos = 0;
    size_t size = str.size();
    State state = State::Before;
    vector<SExpr> stack;
    stack.emplace_back();
    while (pos < size)
    {
        char ch = str[pos++];
        if (state == State::InQuote)
        {
            if (ch == '\n')
            {
                if (pMsg) *pMsg = strprintf("%" PRIuZ ": newline not expected", pos);
                return false;
            }
            stack.back().m_symbol.push_back(ch);
            if (ch == '"')
            {
                state = State::After;
            }
        }
        else if (ch == '(')
        {
            if (state != State::Before)
            {
                if (pMsg) *pMsg = strprintf("%" PRIuZ ": '(' not expected", pos);
                return false;
            }
            stack.back().m_type = SExprType::List;
            stack.emplace_back();
            state = State::Before;
        }
        else if (ch == ')')
        {
            if (stack.size() <= 1)
            {
                if (pMsg) *pMsg = strprintf("%" PRIuZ ": ')' not expected", pos);
                return false;
            }
            if (state != State::Before)
            {
                stack[stack.size() - 2].m_list.push_back(
                    std::move(stack.back()));
            }
            stack.pop_back();
            state = State::After;
        }
        else if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
        {
            if (state != State::Before)
            {
                if (stack.size() <= 1)
                {
                    state = State::After;
                }
                else
                {
                    stack[stack.size() - 2].m_list.push_back(
                        std::move(stack.back()));
                    stack.pop_back();
                    stack.emplace_back();
                    state = State::Before;
                }
            }
        }
        else if (ch == '"')
        {
            if (state != State::Before)
            {
                if (pMsg) *pMsg = strprintf("%" PRIuZ ": '\"' not expected", pos);
                return false;
            }
            stack.back().m_type = SExprType::Symbol;
            state = State::InQuote;
            stack.back().m_symbol.push_back(ch);
        }
        else
        {
            if (state == State::After)
            {
                if (pMsg) *pMsg = strprintf("%" PRIuZ ": symbol not expected", pos);
                return false;
            }
            if (state == State::Before)
            {
                stack.back().m_type = SExprType::Symbol;
                state = State::During;
            }
            stack.back().m_symbol.push_back(ch);
        }
    }

    if (stack.size() != 1)
    {
        if (pMsg) *pMsg = strprintf("%" PRIuZ ": expression not finished", pos);
        return false;
    }
    if (state == State::Before)
    {
        if (pMsg) *pMsg = strprintf("%" PRIuZ ": no expression", pos);
        return false;
    }

    *pExpr = std::move(stack.back());
    return true;
}
