#include "ThreadUtils.hpp"
#include "TimeUtils.hpp"
#include <exception>
#include <stdexcept>

#if PLATFORM_WINDOWS
#include <windows.h>
#else
#include <pthread.h>
#include <unistd.h>
#include <time.h>
#endif

class ThreadImpl
{
public:
    bool m_joinable;
#if PLATFORM_WINDOWS
    HANDLE m_hThread;
#else
    pthread_t m_threadId;
#endif
};

Thread::Thread() :
    m_pImpl(new ThreadImpl)
{
    m_pImpl->m_joinable = false;
}

namespace
{
#if PLATFORM_WINDOWS
    DWORD WINAPI threadFunc(LPVOID arg)
    {
        std::unique_ptr<std::function<void()> >
            pf((std::function<void()>*)arg);
        (*pf)();
        return 0;
    }
#else
    void* threadFunc(void* arg)
    {
        std::unique_ptr<std::function<void()> >
            pf((std::function<void()>*)arg);
        (*pf)();
        return nullptr;
    }
#endif
}

Thread::Thread(std::function<void()> f) :
    m_pImpl(new ThreadImpl)
{
    std::function<void()>* pf = new std::function<void()>(std::move(f));
#if PLATFORM_WINDOWS
    m_pImpl->m_hThread = CreateThread(nullptr,
                                      0,
                                      threadFunc,
                                      pf,
                                      0,
                                      nullptr);
    if (m_pImpl->m_hThread != NULL)
    {
        m_pImpl->m_joinable = true;
    }
    else
    {
        delete pf;
        m_pImpl->m_joinable = false;
    }
#else
    if (pthread_create(&m_pImpl->m_threadId,
                       nullptr,
                       threadFunc,
                       pf) == 0)
    {
        m_pImpl->m_joinable = true;
    }
    else
    {
        delete pf;
        m_pImpl->m_joinable = false;
    }
#endif
}

Thread::Thread(Thread&& other) :
    m_pImpl(new ThreadImpl)
{
    m_pImpl->m_joinable = other.m_pImpl->m_joinable;
#if PLATFORM_WINDOWS
    m_pImpl->m_hThread = other.m_pImpl->m_hThread;
#else
    m_pImpl->m_threadId = other.m_pImpl->m_threadId;
#endif
    other.m_pImpl->m_joinable = false;
}

Thread::~Thread()
{
    if (m_pImpl->m_joinable)
    {
        std::terminate();
    }
}

Thread& Thread::operator=(Thread&& other)
{
    if (m_pImpl->m_joinable)
    {
        std::terminate();
    }
    m_pImpl->m_joinable = other.m_pImpl->m_joinable;
#if PLATFORM_WINDOWS
    m_pImpl->m_hThread = other.m_pImpl->m_hThread;
#else
    m_pImpl->m_threadId = other.m_pImpl->m_threadId;
#endif
    other.m_pImpl->m_joinable = false;

    return *this;
}

void Thread::swap(Thread& other)
{
    using std::swap;
    swap(m_pImpl, other.m_pImpl);
}

bool Thread::joinable() const
{
    return m_pImpl->m_joinable;
}

void Thread::join()
{
    if (!m_pImpl->m_joinable)
    {
        throw std::invalid_argument("this");
    }
#if PLATFORM_WINDOWS
    WaitForSingleObject(m_pImpl->m_hThread, INFINITE);
    CloseHandle(m_pImpl->m_hThread);
#else
    pthread_join(m_pImpl->m_threadId, nullptr);
#endif
    m_pImpl->m_joinable = false;
}

void Thread::detach()
{
    if (!m_pImpl->m_joinable)
    {
        throw std::invalid_argument("this");
    }
#if PLATFORM_WINDOWS
    CloseHandle(m_pImpl->m_hThread);
#else
    pthread_detach(m_pImpl->m_threadId);
#endif
    m_pImpl->m_joinable = false;
}

uint32_t Thread::hardwareConcurrency()
{
#if PLATFORM_WINDOWS
    SYSTEM_INFO systemInfo;
    GetSystemInfo(&systemInfo);
    DWORD result = systemInfo.dwNumberOfProcessors;
    if (result <= 0 || result > 256)
    {
        return 0;
    }
    else
    {
        return (uint32_t)result;
    }
#else
    long result = sysconf(_SC_NPROCESSORS_ONLN);
    if (result <= 0 || result > 256)
    {
        return 0;
    }
    else
    {
        return (uint32_t)result;
    }
#endif
}

void swap(Thread& x, Thread& y)
{
    x.swap(y);
}

class MutexImpl
{
public:
#if PLATFORM_WINDOWS
    CRITICAL_SECTION m_cs;
#else
    pthread_mutex_t m_mutex;
#endif
};

Mutex::Mutex() :
    m_pImpl(new MutexImpl)
{
#if PLATFORM_WINDOWS
    InitializeCriticalSection(&m_pImpl->m_cs);
#else
    if (pthread_mutex_init(&m_pImpl->m_mutex, nullptr) != 0)
    {
        fprintf(stderr, "pthread_mutex_init failed\n");
        exit(1);
    }
#endif
}

Mutex::~Mutex()
{
#if PLATFORM_WINDOWS
    DeleteCriticalSection(&m_pImpl->m_cs);
#else
    pthread_mutex_destroy(&m_pImpl->m_mutex);
#endif
}

void Mutex::lock()
{
#if PLATFORM_WINDOWS
    EnterCriticalSection(&m_pImpl->m_cs);
#else
    if (pthread_mutex_lock(&m_pImpl->m_mutex) != 0)
    {
        fprintf(stderr, "pthread_mutex_lock failed\n");
        exit(1);
    }
#endif
}

bool Mutex::tryLock()
{
#if PLATFORM_WINDOWS
    return TryEnterCriticalSection(&m_pImpl->m_cs) ? true : false;
#else
    int result = pthread_mutex_trylock(&m_pImpl->m_mutex);
    if (result == 0)
    {
        return true;
    }
    else
    {
        if (result != EBUSY)
        {
            fprintf(stderr, "pthread_mutex_trylock failed\n");
            exit(1);
        }
        return false;
    }
#endif
}

void Mutex::unlock()
{
#if PLATFORM_WINDOWS
    LeaveCriticalSection(&m_pImpl->m_cs);
#else
    if (pthread_mutex_unlock(&m_pImpl->m_mutex) != 0)
    {
        fprintf(stderr, "pthread_mutex_lock failed\n");
        exit(1);
    }
#endif
}

class ConditionVariableImpl
{
public:
#if PLATFORM_WINDOWS
    CONDITION_VARIABLE m_cv;
#else
    pthread_cond_t m_cond;
    clockid_t m_clockId;
#endif
};

ConditionVariable::ConditionVariable() :
    m_pImpl(new ConditionVariableImpl)
{
#if PLATFORM_WINDOWS
    InitializeConditionVariable(&m_pImpl->m_cv);
#else
    pthread_condattr_t attr;
    if (pthread_condattr_init(&attr) != 0)
    {
        fprintf(stderr, "pthread_condattr_init failed\n");
        exit(1);
    }
    if (pthread_condattr_setclock(&attr, CLOCK_MONOTONIC_RAW) != 0)
    {
        if (pthread_condattr_setclock(&attr, CLOCK_MONOTONIC) != 0)
        {
            // Use the default clock
        }
    }
    if (pthread_condattr_getclock(&attr, &m_pImpl->m_clockId) != 0)
    {
        fprintf(stderr, "pthread_condattr_getclock failed\n");
        exit(1);
    }
    if (pthread_cond_init(&m_pImpl->m_cond, &attr) != 0)
    {
        fprintf(stderr, "pthread_cond_init failed\n");
        exit(1);
    }
    if (pthread_condattr_destroy(&attr) != 0)
    {
        fprintf(stderr, "pthread_condattr_destroy failed\n");
        exit(1);
    }
#endif
}

ConditionVariable::~ConditionVariable()
{
#if PLATFORM_WINDOWS
#else
    pthread_cond_destroy(&m_pImpl->m_cond);
#endif
}

void ConditionVariable::notifyOne()
{
#if PLATFORM_WINDOWS
    WakeConditionVariable(&m_pImpl->m_cv);
#else
    if (pthread_cond_signal(&m_pImpl->m_cond) != 0)
    {
        fprintf(stderr, "pthread_cond_signal failed\n");
        exit(1);
    }
#endif
}

void ConditionVariable::notifyAll()
{
#if PLATFORM_WINDOWS
    WakeAllConditionVariable(&m_pImpl->m_cv);
#else
    if (pthread_cond_broadcast(&m_pImpl->m_cond) != 0)
    {
        fprintf(stderr, "pthread_cond_broadcast failed\n");
        exit(1);
    }
#endif
}

void ConditionVariable::wait(UniqueLock<Mutex>& lock)
{
#if PLATFORM_WINDOWS
    if (SleepConditionVariableCS(&m_pImpl->m_cv,
                                 &lock.mutex()->m_pImpl->m_cs,
                                 INFINITE) == 0)
    {
        fprintf(stderr, "SleepConditionVariableCS failed\n");
        exit(1);
    }
#else
    if (pthread_cond_wait(&m_pImpl->m_cond,
                          &lock.mutex()->m_pImpl->m_mutex) != 0)
    {
        fprintf(stderr, "pthread_cond_wait failed\n");
        exit(1);
    }
#endif
}

CVStatus ConditionVariable::waitUntilMS(UniqueLock<Mutex>& lock,
                                        uint64_t timeMS)
{
    uint64_t durationMS = timeMS - getTimeMS();
    if (durationMS >= UINT64_C(0x8000000000000000))
    {
        durationMS = 0;
    }
    else if (durationMS > UINT64_C(0x0000000070000000))
    {
        durationMS = UINT64_C(0x0000000070000000);
    }

    return waitForMS(lock, (uint32_t)durationMS);
}

CVStatus ConditionVariable::waitForMS(UniqueLock<Mutex>& lock,
                                      uint32_t durationMS)
{
    if (durationMS > 0x70000000)
    {
        durationMS = 0x70000000;
    }

#if PLATFORM_WINDOWS
    if (SleepConditionVariableCS(&m_pImpl->m_cv,
                                 &lock.mutex()->m_pImpl->m_cs,
                                 (DWORD)durationMS) == 0)
    {
        if (GetLastError() == ERROR_TIMEOUT)
        {
            return CVStatus::Timeout;
        }
        else
        {
            fprintf(stderr, "SleepConditionVariableCS failed\n");
            exit(1);
        }
    }
    else
    {
        return CVStatus::NoTimeout;
    }
#else
    struct timespec ts;
    memset(&ts, 0, sizeof(ts));
    if (clock_gettime(m_pImpl->m_clockId, &ts) != 0)
    {
        fprintf(stderr, "clock_gettime failed\n");
        exit(1);
    }

    ts.tv_sec += (time_t)(durationMS / 1000);
    ts.tv_nsec += (long)((durationMS % 1000) * 1000000);
    if (ts.tv_nsec >= 1000000000)
    {
        ts.tv_nsec -= 1000000000;
        ts.tv_sec++;
    }

    int result =
        pthread_cond_timedwait(&m_pImpl->m_cond,
                               &lock.mutex()->m_pImpl->m_mutex,
                               &ts);
    if (result != 0)
    {
        if (result == ETIMEDOUT)
        {
            return CVStatus::Timeout;
        }
        else
        {
            fprintf(stderr, "pthread_cond_timedwait failed\n");
            exit(1);
        }
    }
    else
    {
        return CVStatus::NoTimeout;
    }
#endif
}

uint64_t ConditionVariable::getTimeMS()
{
    return ::getTimeMS();
}
