#ifndef THREADUTILS_HPP
#define THREADUTILS_HPP

#include "Common.hpp"

class ThreadImpl;

class Thread
{
public:
    Thread();
    explicit Thread(std::function<void()> f);
    Thread(Thread&& other);
    ~Thread();
    Thread& operator=(Thread&& other);

    void swap(Thread& other);
    bool joinable() const;
    void join();
    void detach();

    static uint32_t hardwareConcurrency();

private:
    std::unique_ptr<ThreadImpl> m_pImpl;

private:
    Thread(const Thread& other) = delete;
    Thread& operator=(const Thread& other) = delete;
};

void swap(Thread& x, Thread& y);

class MutexImpl;

class Mutex
{
public:
    Mutex();
    ~Mutex();

    void lock();
    bool tryLock();
    void unlock();

private:
    friend class ConditionVariable;

    std::unique_ptr<MutexImpl> m_pImpl;

private:
    Mutex(const Mutex& other) = delete;
    Mutex(Mutex&& other) = delete;
    Mutex& operator=(const Mutex& other) = delete;
    Mutex& operator=(Mutex&& other) = delete;
};

struct DeferLockType {};
struct TryToLockType {};
struct AdoptLockType {};

constexpr DeferLockType DeferLock{};
constexpr TryToLockType TryToLock{};
constexpr AdoptLockType AdoptLock{};

template<class _MutexType>
class LockGuard
{
public:
    typedef _MutexType MutexType;

    explicit LockGuard(MutexType& mutex) : m_mutex(mutex) { m_mutex.lock(); }
    LockGuard(MutexType& mutex, AdoptLockType) : m_mutex(mutex) {}
    ~LockGuard() { m_mutex.unlock(); }

private:
    MutexType& m_mutex;

private:
    LockGuard(const LockGuard& other) = delete;
    LockGuard(LockGuard&& other) = delete;
    LockGuard& operator=(const LockGuard& other) = delete;
    LockGuard& operator=(LockGuard&& other) = delete;
};

template<class _MutexType>
class UniqueLock
{
public:
    typedef _MutexType MutexType;

    UniqueLock() :
        m_pMutex(nullptr), m_owns(false) {}

    explicit UniqueLock(MutexType& mutex) :
        m_pMutex(&mutex), m_owns(true)
    {
        mutex.lock();
    }

    UniqueLock(MutexType& mutex, DeferLockType) :
        m_pMutex(&mutex), m_owns(false) {}

    UniqueLock(MutexType& mutex, TryToLockType) :
        m_pMutex(&mutex), m_owns(mutex.tryLock()) {}

    UniqueLock(MutexType& mutex, AdoptLockType) :
        m_pMutex(&mutex), m_owns(true) {}

    UniqueLock(UniqueLock&& other) :
        m_pMutex(other.m_pMutex), m_owns(other.m_owns)
    {
        other.m_pMutex = nullptr;
        other.m_owns = false;
    }

    ~UniqueLock()
    {
        if (m_owns)
        {
            m_pMutex->unlock();
        }
    }

    UniqueLock& operator=(UniqueLock&& other)
    {
        if (this != &other)
        {
            if (m_owns)
            {
                m_pMutex->unlock();
            }
            m_pMutex = other.m_pMutex;
            m_owns = other.m_owns;
            other.m_pMutex = nullptr;
            other.m_owns = false;
        }
        return *this;
    }

    void lock()
    {
        m_pMutex->lock();
        m_owns = true;
    }

    bool tryLock()
    {
        m_owns = m_pMutex->tryLock();
        return m_owns;
    }

    void unlock()
    {
        m_pMutex->unlock();
        m_owns = false;
    }

    void swap(UniqueLock& other)
    {
        using std::swap;
        swap(m_pMutex, other.m_pMutex);
        swap(m_owns, other.m_owns);
    }

    MutexType* release()
    {
        MutexType* pMutex = m_pMutex;
        m_pMutex = nullptr;
        m_owns = false;
        return pMutex;
    }

    bool ownsLock() const
    {
        return m_owns;
    }

    explicit operator bool() const
    {
        return m_owns;
    }

    MutexType* mutex() const
    {
        return m_pMutex;
    }

private:
    MutexType* m_pMutex;
    bool m_owns;

private:
    UniqueLock(const UniqueLock& other) = delete;
    UniqueLock& operator=(const UniqueLock& other) = delete;
};

enum class CVStatus : uint32_t
{
    NoTimeout,
    Timeout
};

class ConditionVariableImpl;

class ConditionVariable
{
public:
    ConditionVariable();
    ~ConditionVariable();

    void notifyOne();
    void notifyAll();

    void wait(UniqueLock<Mutex>& lock);

    template<class Predicate>
    void wait(UniqueLock<Mutex>& lock, Predicate pred)
    {
        while (!pred())
        {
            wait(lock);
        }
    }

    CVStatus waitUntilMS(UniqueLock<Mutex>& lock, uint64_t timeMS);

    template<class Predicate>
    bool waitUntilMS(UniqueLock<Mutex>& lock, uint64_t timeMS, Predicate pred)
    {
        while (!pred())
        {
            if (waitUntilMS(lock, timeMS) == CVStatus::Timeout)
            {
                return pred();
            }
        }
        return true;
    }

    CVStatus waitForMS(UniqueLock<Mutex>& lock, uint32_t durationMS);

    template<class Predicate>
    bool waitForMS(UniqueLock<Mutex>& lock, uint32_t durationMS, Predicate pred)
    {
        return waitUntilMS(lock, getTimeMS() + durationMS, std::move(pred));
    }

private:
    static uint64_t getTimeMS();

    std::unique_ptr<ConditionVariableImpl> m_pImpl;

private:
    ConditionVariable(const ConditionVariable& other) = delete;
    ConditionVariable(ConditionVariable&& other) = delete;
    ConditionVariable& operator=(const ConditionVariable& other) = delete;
    ConditionVariable& operator=(ConditionVariable&& other) = delete;
};

enum class FutureStatus : uint32_t
{
    Ready,
    Timeout
};

template<class T> class Future;
template<class T> class Promise;

template<class T>
class SharedStateImpl
{
private:
    friend class Future<T>;
    friend class Promise<T>;

    SharedStateImpl() :
        m_pValue(),
        m_mutex(),
        m_cond(),
        m_refCount(1),
        m_futureRetrieved(false),
        m_valueSet(false)
    {}

    void release()
    {
        bool destroy = false;
        {
            LockGuard<Mutex> lock(m_mutex);
            m_refCount--;
            if (m_refCount == 0)
            {
                destroy = true;
            }
        }
        if (destroy)
        {
            delete this;
        }
    }

    std::unique_ptr<T> m_pValue;
    Mutex m_mutex;
    ConditionVariable m_cond;
    uint32_t m_refCount;
    bool m_futureRetrieved;
    bool m_valueSet;
};

template<class T>
class Promise
{
public:
    Promise() : m_pImpl(new SharedStateImpl<T>) {}

    Promise(Promise&& other) : m_pImpl(other.m_pImpl)
    {
        other.m_pImpl = nullptr;
    }

    ~Promise()
    {
        if (m_pImpl)
        {
            m_pImpl->release();
        }
    }

    Promise& operator=(Promise&& other)
    {
        if (this != &other)
        {
            if (m_pImpl)
            {
                m_pImpl->release();
            }
            m_pImpl = other.m_pImpl;
            other.m_pImpl = nullptr;
        }
        return *this;
    }

    void swap(Promise& other)
    {
        using std::swap;
        swap(m_pImpl, other.m_pImpl);
    }

    Future<T> getFuture()
    {
        if (!m_pImpl)
        {
            fprintf(stderr, "Promise::getFuture: No state\n");
            exit(1);
        }
        LockGuard<Mutex> lock(m_pImpl->m_mutex);
        if (m_pImpl->m_futureRetrieved)
        {
            fprintf(stderr, "Promise::getFuture: Future already retrieved\n");
            exit(1);
        }
        m_pImpl->m_futureRetrieved = true;
        m_pImpl->m_refCount++;
        return Future<T>(m_pImpl);
    }

    void setValue(T value)
    {
        if (!m_pImpl)
        {
            fprintf(stderr, "Promise::setValue: No state\n");
            exit(1);
        }
        LockGuard<Mutex> lock(m_pImpl->m_mutex);
        if (m_pImpl->m_valueSet)
        {
            fprintf(stderr, "Promise::setValue: Promise already satisfied\n");
            exit(1);
        }
        m_pImpl->m_valueSet = true;
        m_pImpl->m_pValue.reset(new T(std::move(value)));
        m_pImpl->m_cond.notifyOne();
    }

private:
    SharedStateImpl<T>* m_pImpl;

private:
    Promise(const Promise& other) = delete;
    Promise& operator=(const Promise& other) = delete;
};

template<class T>
void swap(Promise<T>& x, Promise<T>& y)
{
    x.swap(y);
}

template<class T>
class Future
{
public:
    Future() : m_pImpl(nullptr) {}

    Future(Future&& other) : m_pImpl(other.m_pImpl)
    {
        other.m_pImpl = nullptr;
    }

    ~Future()
    {
        if (m_pImpl)
        {
            m_pImpl->release();
        }
    }

    Future& operator=(Future&& other)
    {
        if (this != &other)
        {
            if (m_pImpl)
            {
                m_pImpl->release();
            }
            m_pImpl = other.m_pImpl;
            other.m_pImpl = nullptr;
        }

        return *this;
    }

    T get()
    {
        std::unique_ptr<T> pValue;
        {
            UniqueLock<Mutex> lock(m_pImpl->m_mutex);
            auto pred = [&]() { return m_pImpl->m_valueSet; };
            m_pImpl->m_cond.wait(lock, pred);
            pValue = std::move(m_pImpl->m_pValue);
        }
        m_pImpl->release();
        m_pImpl = nullptr;
        return std::move(*pValue);
    }

    bool valid() const
    {
        return m_pImpl != nullptr;
    }

    void wait() const
    {
        UniqueLock<Mutex> lock(m_pImpl->m_mutex);
        auto pred = [&]() { return m_pImpl->m_valueSet; };
        m_pImpl->m_cond.wait(lock, pred);
    }

    FutureStatus waitForMS(uint32_t durationMS) const
    {
        UniqueLock<Mutex> lock(m_pImpl->m_mutex);
        auto pred = [&]() { return m_pImpl->m_valueSet; };
        return m_pImpl->m_cond.waitForMS(lock, durationMS, pred) ?
            FutureStatus::Ready :
            FutureStatus::Timeout;
    }

    FutureStatus waitUntilMS(uint64_t timeMS) const
    {
        UniqueLock<Mutex> lock(m_pImpl->m_mutex);
        auto pred = [&]() { return m_pImpl->m_valueSet; };
        return m_pImpl->m_cond.waitUntilMS(lock, timeMS, pred) ?
            FutureStatus::Ready :
            FutureStatus::Timeout;
    }

private:
    friend class Promise<T>;

    Future(SharedStateImpl<T>* pImpl) : m_pImpl(pImpl) {}

    SharedStateImpl<T>* m_pImpl;

private:
    Future(const Future& other) = delete;
    Future& operator=(const Future& other) = delete;
};

#endif
