#include "Value.hpp"
#include "StringUtils.hpp"
#include <limits>

using std::string;
using std::vector;
using std::map;
using std::unique_ptr;

bool decodeInt(const string& tokenStr,
               size_t pos,
               int64_t* pValue,
               string* pMsg)
{
    int64_t value = 0;
    size_t tokenSize = tokenStr.size();
    for (size_t i = pos; i < tokenSize; i++)
    {
        if (value > std::numeric_limits<int64_t>::max() / 94)
        {
            if (pMsg) *pMsg = "Integer literal too large";
            return false;
        }
        value *= 94;
        int64_t digit = tokenStr[i] - (char)33;
        if (digit > std::numeric_limits<int64_t>::max() - value)
        {
            if (pMsg) *pMsg = "Integer literal too large";
            return false;
        }
        value += digit;
    }
    if (pValue) *pValue = value;
    return true;
}

bool decodeInt(const string& tokenStr,
               size_t pos,
               mpz_class* pValue,
               string* pMsg)
{
    mpz_class& value = *pValue;
    value = 0;
    size_t tokenSize = tokenStr.size();
    for (size_t i = pos; i < tokenSize; i++)
    {
        value *= 94;
        uint32_t digit = tokenStr[i] - (char)33;
        value += digit;
    }
    return true;
}

bool encodeInt(int64_t value,
               string& intStr,
               string* pMsg)
{
    if (value < 0)
    {
        if (pMsg) *pMsg = "Cannot encode negative integer";
        return false;
    }
    if (value == 0)
    {
        intStr += (char)33;
        return true;
    }
    size_t pos = intStr.size();
    while (value != 0)
    {
        intStr += (char)(33 + value % 94);
        value /= 94;
    }
    std::reverse(intStr.begin() + pos, intStr.end());
    return true;
}

bool encodeInt(const mpz_class& value,
               string& intStr,
               string* pMsg)
{
    if (value < 0)
    {
        if (pMsg) *pMsg = "Cannot encode negative integer";
        return false;
    }
    if (value == 0)
    {
        intStr += (char)33;
        return true;
    }
    size_t pos = intStr.size();
    mpz_class q(value);
    while (q != 0)
    {
        unsigned long r = mpz_tdiv_q_ui(q.get_mpz_t(), q.get_mpz_t(), 94);
        intStr += (char)(33 + r);
    }
    std::reverse(intStr.begin() + pos, intStr.end());
    return true;
}

bool decodeString(const string& strStr,
                  size_t pos,
                  string* pStr,
                  string* pMsg)
{
    string str;
    for (size_t i = pos; i < strStr.size(); i++)
    {
        char tokenCh = strStr[i];
        if (tokenCh < (char)33 || tokenCh > (char)126)
        {
            if (pMsg) *pMsg = "Invalid character in token";
            return false;
        }
        size_t tokenVal = tokenCh - (char)33;
        char ch = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!\"#$%&'()*+,-./:;<=>?@[\\]^_`|~ \n"[tokenVal];
        str += ch;
    }
    if (pStr) *pStr = std::move(str);
    return true;
}

namespace
{
    char encodeCharMap[256] =
        "\0\0\0\0\0\0\0\0\0\0~\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
        "}_`abcdefghijklmUVWXYZ[\\]^nopqrs"
        "t;<=>?@ABCDEFGHIJKLMNOPQRSTuvwxy"
        "z!\"#$%&'()*+,-./0123456789:\0{\0|\0"
        "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
        "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
        "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
        "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
}

bool encodeString(const string& str,
                  string& strStr,
                  string* pMsg)
{
    for (char ch : str)
    {
        char encodedCh = encodeCharMap[(uint8_t)ch];
        if (encodedCh == '\0')
        {
            if (pMsg) *pMsg = "Invalid character in string value";
            return false;
        }
        strStr += encodedCh;
    }
    return true;
}

namespace
{
    vector<string> split(const string& valueStr)
    {
        vector<string> tokenStrs;
        size_t size = valueStr.size();
        size_t pos = 0;
        while (true)
        {
            while (pos < size && valueStr[pos] == ' ') pos++;
            if (pos == size) break;
            string& tokenStr = tokenStrs.emplace_back();
            while (pos < size && valueStr[pos] != ' ') tokenStr += valueStr[pos++];
        }
        return tokenStrs;
    }

    bool decode(const vector<string>& tokenStrs,
                size_t& pos,
                map<int64_t, int64_t>& varIdToIdx,
                int64_t varCount,
                Value* pValue,
                string* pMsg = nullptr)
    {
        size_t size = tokenStrs.size();
        if (pos >= size)
        {
            if (pMsg) *pMsg = "Unexpected end of input";
            return false;
        }

        auto& tokenStr = tokenStrs[pos++];
        if (tokenStr.empty())
        {
            if (pMsg) *pMsg = "Empty token";
            return false;
        }

        for (char ch : tokenStr)
        {
            if (ch < (char)33 || ch > (char)126)
            {
                if (pMsg) *pMsg = "Invalid character in token";
                return false;
            }
        }

        Value value;
        value.init();

        size_t tokenSize = tokenStr.size();
        char tokenType = tokenStr[0];
        switch (tokenType)
        {
        case 'T':
        {
            if (tokenSize != 1)
            {
                if (pMsg) *pMsg = "Unexpected body after 'T'";
                return false;
            }
            value->setValueType(ValueType::Bool);
            value->boolData.boolValue = true;
            break;
        }
        case 'F':
        {
            if (tokenSize != 1)
            {
                if (pMsg) *pMsg = "Unexpected body after 'F'";
                return false;
            }
            value->setValueType(ValueType::Bool);
            value->boolData.boolValue = false;
            break;
        }
        case 'I':
        {
            value->setValueType(ValueType::Int);
            if (!decodeInt(tokenStr, 1, &value->intData.intValue, pMsg)) return false;
            break;
        }
        case 'S':
        {
            value->setValueType(ValueType::String);
            if (!decodeString(tokenStr, 1, &value->stringData.stringValue, pMsg)) return false;
            break;
        }
        case 'U':
        {
            if (tokenSize != 2)
            {
                if (pMsg) *pMsg = "Unexpected unary operator size";
                return false;
            }
            value->setValueType(ValueType::Unary);
            switch (tokenStr[1])
            {
            case '-': value->unaryData.unaryOp = UnaryOp::Negate; break;
            case '!': value->unaryData.unaryOp = UnaryOp::Not; break;
            case '#': value->unaryData.unaryOp = UnaryOp::ToInt; break;
            case '$': value->unaryData.unaryOp = UnaryOp::ToString; break;
            default:
            {
                if (pMsg) *pMsg = "Unknown unary operator";
                return false;
            }
            }
            if (!decode(tokenStrs, pos, varIdToIdx, varCount, &value->unaryData.arg, pMsg))
            {
                return false;
            }
            break;
        }
        case 'B':
        {
            if (tokenSize != 2)
            {
                if (pMsg) *pMsg = "Unexpected binary operator size";
                return false;
            }
            value->setValueType(ValueType::Binary);
            switch (tokenStr[1])
            {
            case '+': value->binaryData.binaryOp = BinaryOp::Add; break;
            case '-': value->binaryData.binaryOp = BinaryOp::Sub; break;
            case '*': value->binaryData.binaryOp = BinaryOp::Mul; break;
            case '/': value->binaryData.binaryOp = BinaryOp::Div; break;
            case '%': value->binaryData.binaryOp = BinaryOp::Mod; break;
            case '<': value->binaryData.binaryOp = BinaryOp::Lt; break;
            case '>': value->binaryData.binaryOp = BinaryOp::Gt; break;
            case '=': value->binaryData.binaryOp = BinaryOp::Eq; break;
            case '|': value->binaryData.binaryOp = BinaryOp::Or; break;
            case '&': value->binaryData.binaryOp = BinaryOp::And; break;
            case '.': value->binaryData.binaryOp = BinaryOp::Concat; break;
            case 'T': value->binaryData.binaryOp = BinaryOp::Take; break;
            case 'D': value->binaryData.binaryOp = BinaryOp::Drop; break;
            case '$': value->binaryData.binaryOp = BinaryOp::Apply; break;
            default:
            {
                if (pMsg) *pMsg = "Unknown binary operator";
                return false;
            }
            }
            if (!decode(tokenStrs, pos, varIdToIdx, varCount, &value->binaryData.args[0], pMsg) ||
                !decode(tokenStrs, pos, varIdToIdx, varCount, &value->binaryData.args[1], pMsg))
            {
                return false;
            }
            break;
        }
        case '?':
        {
            if (tokenSize != 1)
            {
                if (pMsg) *pMsg = "Unexpected body after '?'";
                return false;
            }
            value->setValueType(ValueType::If);
            if (!decode(tokenStrs, pos, varIdToIdx, varCount, &value->ifData.condValue, pMsg) ||
                !decode(tokenStrs, pos, varIdToIdx, varCount, &value->ifData.trueValue, pMsg) ||
                !decode(tokenStrs, pos, varIdToIdx, varCount, &value->ifData.falseValue, pMsg))
            {
                return false;
            }
            break;
        }
        case 'L':
        {
            value->setValueType(ValueType::Lambda);
            int64_t varId;
            if (!decodeInt(tokenStr, 1, &varId, pMsg)) return false;
            int64_t oldVarIdx = -1;
            auto findIt = varIdToIdx.find(varId);
            if (findIt != varIdToIdx.end())
            {
                oldVarIdx = findIt->second;
            }
            varIdToIdx[varId] = varCount;
            value->varData.var = varCount;
            if (!decode(tokenStrs, pos, varIdToIdx, varCount + 1, &value->lambdaData.body, pMsg))
            {
                return false;
            }
            if (oldVarIdx == -1)
            {
                varIdToIdx.erase(varId);
            }
            else
            {
                varIdToIdx[varId] = oldVarIdx;
            }
            break;
        }
        case 'v':
        {
            value->setValueType(ValueType::Var);
            int64_t varId;
            if (!decodeInt(tokenStr, 1, &varId, pMsg)) return false;
            auto findIt = varIdToIdx.find(varId);
            if (findIt == varIdToIdx.end())
            {
                if (pMsg) *pMsg = "Unbound variable";
                return false;
            }
            value->varData.var = findIt->second;
            break;
        }
        default:
        {
            if (pMsg) *pMsg = "Unsupported token type";
            return false;
        }
        }

        if (pValue) *pValue = std::move(value);
        return true;
    }
}

bool decodeValue(const string& valueStr,
                 Value* pValue,
                 string* pMsg)
{
    vector<string> tokenStrs = split(valueStr);

    size_t pos = 0;
    map<int64_t, int64_t> varIdToIdx;

    Value value;
    if (!decode(tokenStrs, pos, varIdToIdx, 0, &value, pMsg)) return false;
    if (pos != tokenStrs.size())
    {
        if (pMsg) *pMsg = "Unexpected extra input";
        return false;
    }
    if (pValue) *pValue = std::move(value);
    return true;
}

namespace
{
    bool encodeValueImpl(const Value& value,
                         string& valueStr,
                         string* pMsg)
    {
        if (!value)
        {
            if (pMsg) *pMsg = "Null value";
            return false;
        }

        switch (value->valueType)
        {
        case ValueType::Invalid:
        {
            if (pMsg) *pMsg = "Invalid value";
            return false;
        }
        case ValueType::Bool:
        {
            valueStr += value->boolData.boolValue ? 'T' : 'F';
            break;
        }
        case ValueType::Int:
        {
            valueStr += 'I';
            if (!encodeInt(value->intData.intValue, valueStr, pMsg)) return false;
            break;
        }
        case ValueType::String:
        {
            valueStr += 'S';
            if (!encodeString(value->stringData.stringValue, valueStr, pMsg)) return false;
            break;
        }
        case ValueType::Unary:
        {
            valueStr += 'U';
            switch (value->unaryData.unaryOp)
            {
            case UnaryOp::Invalid:
                if (pMsg) *pMsg = "Invalid unary operator";
                return false;
            case UnaryOp::Negate: valueStr += '-'; break;
            case UnaryOp::Not: valueStr += '!'; break;
            case UnaryOp::ToInt: valueStr += '#'; break;
            case UnaryOp::ToString: valueStr += '$'; break;
            }
            valueStr += ' ';
            if (!encodeValueImpl(value->unaryData.arg, valueStr, pMsg)) return false;
            break;
        }
        case ValueType::Binary:
        {
            valueStr += 'B';
            switch (value->binaryData.binaryOp)
            {
            case BinaryOp::Invalid:
                if (pMsg) *pMsg = "Invalid binary operator";
                return false;
            case BinaryOp::Add: valueStr += '+'; break;
            case BinaryOp::Sub: valueStr += '-'; break;
            case BinaryOp::Mul: valueStr += '*'; break;
            case BinaryOp::Div: valueStr += '/'; break;
            case BinaryOp::Mod: valueStr += '%'; break;
            case BinaryOp::Lt: valueStr += '<'; break;
            case BinaryOp::Gt: valueStr += '>'; break;
            case BinaryOp::Eq: valueStr += '='; break;
            case BinaryOp::Or: valueStr += '|'; break;
            case BinaryOp::And: valueStr += '&'; break;
            case BinaryOp::Concat: valueStr += '.'; break;
            case BinaryOp::Take: valueStr += 'T'; break;
            case BinaryOp::Drop: valueStr += 'D'; break;
            case BinaryOp::Apply: valueStr += '$'; break;
            }
            valueStr += ' '     ;
            if (!encodeValueImpl(value->binaryData.args[0], valueStr, pMsg)) return false;
            valueStr += ' ';
            if (!encodeValueImpl(value->binaryData.args[1], valueStr, pMsg)) return false;
            break;
        }
        case ValueType::If:
        {
            valueStr += '?';
            valueStr += ' ';
            if (!encodeValueImpl(value->ifData.condValue, valueStr, pMsg)) return false;
            valueStr += ' ';
            if (!encodeValueImpl(value->ifData.trueValue, valueStr, pMsg)) return false;
            valueStr += ' ';
            if (!encodeValueImpl(value->ifData.falseValue, valueStr, pMsg)) return false;
            break;
        }
        case ValueType::Lambda:
        {
            valueStr += 'L';
            if (!encodeInt(value->lambdaData.var, valueStr, pMsg)) return false;
            valueStr += ' ';
            if (!encodeValueImpl(value->lambdaData.body, valueStr, pMsg)) return false;
            break;
        }
        case ValueType::Var:
        {
            valueStr += 'v';
            if (!encodeInt(value->varData.var, valueStr, pMsg)) return false;
            break;
        }
        default:
        {
            if (pMsg) *pMsg = "Unsupported value type";
            return false;
        }
        }

        return true;
    }
}

bool encodeValue(const Value& value,
                 string* pValueStr,
                 string* pMsg)
{
    string valueStr;
    if (!encodeValueImpl(value, valueStr, pMsg)) return false;
    if (pValueStr) *pValueStr = std::move(valueStr);
    return true;
}

namespace
{
    bool formatValueImpl(const Value& value,
                         string& str,
                         string* pMsg)
    {
        if (!value)
        {
            if (pMsg) *pMsg = "Null value";
            return false;
        }

        switch (value->valueType)
        {
        case ValueType::Invalid:
        {
            if (pMsg) *pMsg = "Invalid value";
            return false;
        }
        case ValueType::Bool:
        {
            str += value->boolData.boolValue ? "true" : "false";
            break;
        }
        case ValueType::Int:
        {
            str += value->intData.intValue.get_str();
            break;
        }
        case ValueType::String:
        {
            str += '"';
            for (char ch : value->stringData.stringValue)
            {
                if (ch == '\n')
                {
                    str += "\\n";
                }
                else
                {
                    if (ch == '"' || ch == '\\')
                    {
                        str += '\\';
                    }
                    str += ch;
                }
            }
            str += '"';
            break;
        }
        case ValueType::Unary:
        {
            str += '(';
            switch (value->unaryData.unaryOp)
            {
            case UnaryOp::Invalid:
                if (pMsg) *pMsg = "Invalid unary operator";
                return false;
            case UnaryOp::Negate: str += '-'; break;
            case UnaryOp::Not: str += '!'; break;
            case UnaryOp::ToInt: str += "toInt"; break;
            case UnaryOp::ToString: str += "toString"; break;
            }
            str += ' ';

            if (!formatValueImpl(value->unaryData.arg, str, pMsg))
            {
                return false;
            }

            str += ')';
            break;
        }
        case ValueType::Binary:
        {
            str += '(';
            switch (value->binaryData.binaryOp)
            {
            case BinaryOp::Invalid:
                if (pMsg) *pMsg = "Invalid binary operator";
                return false;
            case BinaryOp::Add: str += '+'; break;
            case BinaryOp::Sub: str += '-'; break;
            case BinaryOp::Mul: str += '*'; break;
            case BinaryOp::Div: str += '/'; break;
            case BinaryOp::Mod: str += '%'; break;
            case BinaryOp::Lt: str += '<'; break;
            case BinaryOp::Gt: str += '>'; break;
            case BinaryOp::Eq: str += '='; break;
            case BinaryOp::Or: str += "or"; break;
            case BinaryOp::And: str += "and"; break;
            case BinaryOp::Concat: str += "concat"; break;
            case BinaryOp::Take: str += "take"; break;
            case BinaryOp::Drop: str += "drop"; break;
            case BinaryOp::Apply: break;
            }

            if (value->binaryData.binaryOp != BinaryOp::Apply)
            {
                str += ' ';
            }

            if (!formatValueImpl(value->binaryData.args[0], str, pMsg)) return false;
            str += ' ';
            if (!formatValueImpl(value->binaryData.args[1], str, pMsg)) return false;

            str += ')';
            break;
        }
        case ValueType::If:
        {
            str += "(if ";

            if (!formatValueImpl(value->ifData.condValue, str, pMsg)) return false;
            str += ' ';
            if (!formatValueImpl(value->ifData.trueValue, str, pMsg)) return false;
            str += ' ';
            if (!formatValueImpl(value->ifData.falseValue, str, pMsg)) return false;

            str += ')';
            break;
        }
        case ValueType::Lambda:
        {
            str += "(\\v";
            str += strprintf("%" PRIi64 "", value->lambdaData.var);
            str += " -> ";
            if (!formatValueImpl(value->lambdaData.body, str, pMsg)) return false;
            str += ')';
            break;
        }
        case ValueType::Var:
        {
            str += "v";
            str += strprintf("%" PRIi64 "", value->varData.var);
            break;
        }
        default:
        {
            if (pMsg) *pMsg = "Unsupported value type";
            return false;
        }
        }

        return true;
    }
}

bool formatValue(const Value& value,
                 string* pStr,
                 string* pMsg)
{
    string str;
    if (!formatValueImpl(value, str, pMsg)) return false;
    if (pStr) *pStr = std::move(str);
    return true;
}
