#ifndef VALUE_HPP
#define VALUE_HPP

#include "Common.hpp"
#include "FixedSizeHeap.hpp"
#include <gmpxx.h>

class Value;
class ValueData;

// ValueType

enum class ValueType : uint32_t
{
    Invalid,
    Bool,
    Int,
    String,
    Unary,
    Binary,
    If,
    Lambda,
    Var,
    Closure,
    Thunk
};

enum class UnaryOp : uint32_t
{
    Invalid,
    Negate,
    Not,
    ToInt,
    ToString
};

enum class BinaryOp : uint32_t
{
    Invalid,
    Add,
    Sub,
    Mul,
    Div,
    Mod,
    Lt,
    Gt,
    Eq,
    Or,
    And,
    Concat,
    Take,
    Drop,
    Apply
};

// Value

class Value
{
public:
    Value();
    Value(const Value& other);
    Value(Value&& other);
    ~Value();
    Value& operator=(const Value& other);
    Value& operator=(Value&& other);
    void swap(Value& other);

    explicit operator bool() const;
    ValueData& operator*() const;
    ValueData* operator->() const;

    void init();
    void init(ValueType valueType);

private:
    ValueData* pData;
};

// ValueData

class ValueBoolData
{
public:
    bool boolValue = false;
};

class ValueIntData
{
public:
    mpz_class intValue;
};

class ValueStringData
{
public:
    std::string stringValue;
};

class ValueUnaryData
{
public:
    UnaryOp unaryOp = UnaryOp::Invalid;
    Value arg;
};

class ValueBinaryData
{
public:
    BinaryOp binaryOp = BinaryOp::Invalid;
    Value args[2];
};

class ValueIfData
{
public:
    Value condValue;
    Value trueValue;
    Value falseValue;
};

class ValueLambdaData
{
public:
    int64_t var = -1;
    Value body;
};

class ValueVarData
{
public:
    int64_t var = -1;
};

class ValueClosureData
{
public:
    std::vector<Value> vars;
    Value body;
};

class ValueThunkData
{
public:
    std::vector<Value> vars;
    Value body;
};

class ValueData
{
public:
    ValueData();
    ValueData(const ValueData& other);
    ValueData(ValueData&& other);
    ~ValueData();
    ValueData& operator=(const ValueData& other);
    ValueData& operator=(ValueData&& other);

    static ValueData* create();
    void addRef();
    void release();

    ValueType getValueType() const;
    void setValueType(ValueType valueType);

private:
    void destroy();
    void copyFrom(const ValueData& other);
    void moveFrom(ValueData&& other);

public:
    ValueType valueType;
    uint32_t refCount;
    union
    {
        ValueBoolData boolData;
        ValueIntData intData;
        ValueStringData stringData;
        ValueUnaryData unaryData;
        ValueBinaryData binaryData;
        ValueIfData ifData;
        ValueLambdaData lambdaData;
        ValueVarData varData;
        ValueClosureData closureData;
        ValueThunkData thunkData;
    };
};

// Value implementation

inline Value::Value() :
    pData(nullptr)
{
}

inline Value::Value(const Value& other) :
    pData(other.pData)
{
    if (pData) pData->addRef();
}

inline Value::Value(Value&& other) :
    pData(other.pData)
{
    other.pData = nullptr;
}

inline Value::~Value()
{
    if (pData) pData->release();
}

inline Value& Value::operator=(const Value& other)
{
    if (this != &other)
    {
        if (other.pData) other.pData->addRef();
        if (pData) pData->release();
        pData = other.pData;
    }
    return *this;
}

inline Value& Value::operator=(Value&& other)
{
    if (this != &other)
    {
        if (pData) pData->release();
        pData = other.pData;
        other.pData = nullptr;
    }
    return *this;
}

inline void Value::swap(Value& other)
{
    using std::swap;
    swap(pData, other.pData);
}

inline Value::operator bool() const
{
    return pData != nullptr;
}

inline ValueData& Value::operator*() const
{
    return *pData;
}

inline ValueData* Value::operator->() const
{
    return pData;
}

inline void Value::init()
{
    if (pData) pData->release();
    pData = ValueData::create();
}

inline void Value::init(ValueType valueType)
{
    init();
    pData->setValueType(valueType);
}

inline void swap(Value& x, Value& y)
{
    x.swap(y);
}

// ValueData implementation

inline FixedSizeHeap<sizeof(ValueData)> valueHeap{(size_t)1 * 1024 * 1024 * 1024};

inline ValueData::ValueData() :
    valueType(ValueType::Invalid),
    refCount(1)
{
}

inline ValueData::ValueData(const ValueData& other)
{
    copyFrom(other);
}

inline ValueData::ValueData(ValueData&& other)
{
    moveFrom(std::move(other));
}

inline ValueData::~ValueData()
{
    destroy();
}

inline ValueData& ValueData::operator=(const ValueData& other)
{
    if (this != &other)
    {
        destroy();
        copyFrom(other);
    }
    return *this;
}

inline ValueData& ValueData::operator=(ValueData&& other)
{
    if (this != &other)
    {
        destroy();
        moveFrom(std::move(other));
    }
    return *this;
}

inline /*static*/ ValueData* ValueData::create()
{
    ValueData* pData = (ValueData*)valueHeap.malloc();
    new(pData) ValueData;
    return pData;
}

inline void ValueData::addRef()
{
    refCount++;
}

inline void ValueData::release()
{
    if (--refCount == 0)
    {
        this->~ValueData();
        valueHeap.free(this);
    }
}

inline ValueType ValueData::getValueType() const
{
    return valueType;
}

inline void ValueData::setValueType(ValueType valueType_)
{
    if (valueType_ != valueType)
    {
        destroy();
        valueType = valueType_;
        switch (valueType_)
        {
        case ValueType::Invalid: break;
        case ValueType::Bool: new(&boolData) ValueBoolData; break;
        case ValueType::Int: new(&intData) ValueIntData; break;
        case ValueType::String: new(&stringData) ValueStringData; break;
        case ValueType::Unary: new(&unaryData) ValueUnaryData; break;
        case ValueType::Binary: new(&binaryData) ValueBinaryData; break;
        case ValueType::If: new(&ifData) ValueIfData; break;
        case ValueType::Lambda: new(&lambdaData) ValueLambdaData; break;
        case ValueType::Var: new(&varData) ValueVarData; break;
        case ValueType::Closure: new(&closureData) ValueClosureData; break;
        case ValueType::Thunk: new(&thunkData) ValueThunkData; break;
        }
    }
}

inline void ValueData::destroy()
{
    switch (valueType)
    {
    case ValueType::Invalid: break;
    case ValueType::Bool: boolData.~ValueBoolData(); break;
    case ValueType::Int: intData.~ValueIntData(); break;
    case ValueType::String: stringData.~ValueStringData(); break;
    case ValueType::Unary: unaryData.~ValueUnaryData(); break;
    case ValueType::Binary: binaryData.~ValueBinaryData(); break;
    case ValueType::If: ifData.~ValueIfData(); break;
    case ValueType::Lambda: lambdaData.~ValueLambdaData(); break;
    case ValueType::Var: varData.~ValueVarData(); break;
    case ValueType::Closure: closureData.~ValueClosureData(); break;
    case ValueType::Thunk: thunkData.~ValueThunkData(); break;
    }
}

inline void ValueData::copyFrom(const ValueData& other)
{
    valueType = other.valueType;

    switch (other.valueType)
    {
    case ValueType::Invalid: break;
    case ValueType::Bool: new(&boolData) ValueBoolData(other.boolData); break;
    case ValueType::Int: new(&intData) ValueIntData(other.intData); break;
    case ValueType::String: new(&stringData) ValueStringData(other.stringData); break;
    case ValueType::Unary: new(&unaryData) ValueUnaryData(other.unaryData); break;
    case ValueType::Binary: new(&binaryData) ValueBinaryData(other.binaryData); break;
    case ValueType::If: new(&ifData) ValueIfData(other.ifData); break;
    case ValueType::Lambda: new(&lambdaData) ValueLambdaData(other.lambdaData); break;
    case ValueType::Var: new(&varData) ValueVarData(other.varData); break;
    case ValueType::Closure: new(&closureData) ValueClosureData(other.closureData); break;
    case ValueType::Thunk: new(&thunkData) ValueThunkData(other.thunkData); break;
    }
}

inline void ValueData::moveFrom(ValueData&& other)
{
    valueType = other.valueType;

    switch (other.valueType)
    {
    case ValueType::Invalid: break;
    case ValueType::Bool: new(&boolData) ValueBoolData(std::move(other.boolData)); break;
    case ValueType::Int: new(&intData) ValueIntData(std::move(other.intData)); break;
    case ValueType::String: new(&stringData) ValueStringData(std::move(other.stringData)); break;
    case ValueType::Unary: new(&unaryData) ValueUnaryData(std::move(other.unaryData)); break;
    case ValueType::Binary: new(&binaryData) ValueBinaryData(std::move(other.binaryData)); break;
    case ValueType::If: new(&ifData) ValueIfData(std::move(other.ifData)); break;
    case ValueType::Lambda: new(&lambdaData) ValueLambdaData(std::move(other.lambdaData)); break;
    case ValueType::Var: new(&varData) ValueVarData(std::move(other.varData)); break;
    case ValueType::Closure: new(&closureData) ValueClosureData(std::move(other.closureData)); break;
    case ValueType::Thunk: new(&thunkData) ValueThunkData(std::move(other.thunkData)); break;
    }
}

bool decodeInt(const std::string& tokenStr,
               size_t pos,
               int64_t* pValue,
               std::string* pMsg = nullptr);

bool decodeInt(const std::string& tokenStr,
               size_t pos,
               mpz_class* pValue,
               std::string* pMsg = nullptr);

bool encodeInt(int64_t value,
               std::string& intStr,
               std::string* pMsg = nullptr);

bool encodeInt(const mpz_class& value,
               std::string& intStr,
               std::string* pMsg = nullptr);

bool decodeString(const std::string& strStr,
                  size_t pos,
                  std::string* pStr,
                  std::string* pMsg = nullptr);

bool encodeString(const std::string& str,
                  std::string& strStr,
                  std::string* pMsg = nullptr);

bool decodeValue(const std::string& valueStr,
                 Value* pValue,
                 std::string* pMsg = nullptr);

bool encodeValue(const Value& value,
                 std::string* pValueStr,
                 std::string* pMsg = nullptr);

bool formatValue(const Value& value,
                 std::string* pStr,
                 std::string* pMsg = nullptr);

#endif
