#include "Common.hpp"
#include "Cleanup.hpp"
#include "Protocol.hpp"
#include "SExpr.hpp"
#include "Value.hpp"
#include "Compile.hpp"
#include "Eval.hpp"

using std::string;
using std::vector;
using std::unique_ptr;

void usage()
{
    printf("Usage: comm [<options>] <string>\n");
    printf("  <string>\n");
    printf("        String to communicate to server\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -r    Treat string as raw encoded data to send\n");
    printf("  -e    Treat string as expression to compile and send\n");
}

int main(int argc, char *argv[])
{
    string str;

    bool help = false;
    bool useRaw = false;
    bool useExpr = false;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-r")
        {
            if (useRaw || useExpr)
            {
                usage();
                return 1;
            }
            useRaw = true;
        }
        else if (strArg == "-e")
        {
            if (useRaw || useExpr)
            {
                usage();
                return 1;
            }
            useExpr = true;
        }
        else
        {
            if (!strArg.empty())
            {
                if (!str.empty()) str += ' ';
                str += strArg;
            }
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (str.empty())
    {
        usage();
        return 1;
    }

    //printf("%s\n", str.c_str());

    string msg;

    string requestStr;
    if (useRaw)
    {
        requestStr = str;
    }
    else if (useExpr)
    {
        SExpr expr;
        if (!parseSExpr(str, &expr, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Value value;
        if (!compile(expr, &value, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        if (!encodeValue(value, &requestStr, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        //printf("%s\n", requestStr.c_str());
    }
    else
    {
        Value value;
        value.init(ValueType::String);
        value->stringData.stringValue = str;

        if (!encodeValue(value, &requestStr, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        //printf("%s\n", requestStr.c_str());
    }

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string responseStr;
    if (!Protocol::communicate(requestStr, &responseStr, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    //printf("%s\n", responseStr.c_str());

    Value responseValue;
    if (!decodeValue(responseStr, &responseValue, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

#if 0
    string encoded;
    if (!encodeValue(responseValue, &encoded, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    if (encoded != responseStr)
    {
        printf("Re-encoded response does not match original\n");
        printf("%s\n", encoded.c_str());
        //return 1;
    }
#endif

#if 0
    string formatted;
    if (!formatValue(responseValue, &formatted, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    printf("%s\n", formatted.c_str());
#endif

    Value evalValue;
    if (!eval(responseValue, &evalValue, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    if (evalValue->valueType == ValueType::String)
    {
        printf("%s\n", evalValue->stringData.stringValue.c_str());
    }
    else
    {
        string evalFormatted;
        if (!formatValue(evalValue, &evalFormatted, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }
        printf("%s\n", evalFormatted.c_str());
    }

    return 0;
}
