#include "Common.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "SExpr.hpp"
#include "Value.hpp"
#include "Compile.hpp"
#include "Eval.hpp"
#include "Cleanup.hpp"
#include "Protocol.hpp"

using std::string;
using std::vector;
using std::map;
using std::pair;

class Sat
{
public:
    class Clause
    {
    public:
        uint64_t used = 0;
        uint64_t values = 0;
    };

    void setBitCount(uint64_t bitCount_)
    {
        bitCount = bitCount_;
    }

    void addClause(uint64_t used, uint64_t values)
    {
        auto& clause = clauses.emplace_back();
        clause.used = used;
        clause.values = values;
    }

    void recurse(uint64_t bitsLeft, uint64_t ans)
    {
        //printf("recurse\n");
        while (true)
        {
            bool didSomething = false;
            for (auto& clause : clauses)
            {
                uint64_t used = clause.used;
                uint64_t values = clause.values;

                //printf("clause  0x%016" PRIx64 " 0x%016" PRIx64 "\n", used, values);
                uint64_t clauseLeft = used & bitsLeft;
                if (clauseLeft == 0)
                {
                    if (((ans & used) ^ values) == used)
                    {
                        // Backtrack
                        //printf("backtrack\n");
                        return;
                    }
                    continue;
                }

                if ((clauseLeft & (clauseLeft - 1)) == 0)
                {
                    uint64_t curUsed = used & ~bitsLeft;
                    if (((ans & curUsed) ^ (values & curUsed)) == curUsed)
                    {
                        bitsLeft &= ~clauseLeft;
                        ans |= (clauseLeft & values);
                        //printf("setting 0x%016" PRIx64 " 0x%016" PRIx64 "\n", bitsLeft, ans);
                        didSomething = true;
                    }
                }
            }
            if (!didSomething) break;
        }
        //printf("!didSomething\n");

        if (bitsLeft == 0)
        {
            bool isMin = false;
            if (ans < minAns)
            {
                isMin = true;
                minAns = ans;
            }
            printf("ans = 0x%016" PRIx64 " %" PRIu64 "%s\n", ans, ans, isMin ? "  new minimum" : "");
            solutionCount++;
            return;
        }

        uint64_t newBitsLeft = bitsLeft & (bitsLeft - 1);
        uint64_t bit = bitsLeft ^ newBitsLeft;
        recurse(newBitsLeft, ans);
        recurse(newBitsLeft, ans | bit);
    }

    void solve()
    {
        uint64_t bitsLeft = ((uint64_t)1 << bitCount) - 1;
        uint64_t ans = 0;
        recurse(bitsLeft, ans);
        printf("solutionCount = %" PRIu64 "\n", solutionCount);
    }

    uint64_t bitCount = 0;
    uint64_t solutionCount = 0;
    uint64_t minAns = (uint64_t)-1;
    vector<Clause> clauses;
};

class Eq
{
public:
    class Clause
    {
    public:
        uint64_t digits[2] = { 0, 0 };
    };

    void setDigitCount(uint64_t digitCount_)
    {
        digitCount = digitCount_;
        ans.clear();
        ans.resize(digitCount, 9);
    }

    void addClause(uint64_t digit0, uint64_t digit1)
    {
        if (digit0 == digit1)
        {
            printf("digit0 == digit1\n");
            exit(1);
        }
        auto& clause = clauses.emplace_back();
        clause.digits[0] = digit0;
        clause.digits[1] = digit1;
    }

    void addDigit(uint64_t digit, uint64_t val)
    {
        if (ans[digit] != 9)
        {
            printf("ans[digit] != 9\n");
            exit(1);
        }
        ans[digit] = val;
    }

    void recurse(uint64_t digitsLeft)
    {
        if (digitsLeft == 0)
        {
            mpz_class big;
            for (uint64_t digit = 0; digit < digitCount; digit++)
            {
                big *= 9;
                big += ans[digit];
            }
            printf("ans = %s\n", big.get_str().c_str());
            solutionCount++;
            return;
        }

        uint64_t bestDigit = (uint64_t)-1;
        //uint64_t bestCount = 10;
        uint64_t bestValid = 0;

        for (uint64_t digit = 0; digit < digitCount; digit++)
        {
            if (ans[digit] != 9) continue;
            uint64_t valid = 0b111111111;
            for (auto& clause : clauses)
            {
                if (clause.digits[0] == digit)
                {
                    uint64_t otherVal = ans[clause.digits[1]];
                    if (otherVal != 9)
                    {
                        valid &= ~((uint64_t)1 << otherVal);
                    }
                }
                if (clause.digits[1] == digit)
                {
                    uint64_t otherVal = ans[clause.digits[0]];
                    if (otherVal != 9)
                    {
                        valid &= ~((uint64_t)1 << otherVal);
                    }
                }
            }
            uint64_t count = __builtin_popcountll(valid);
            if (count == 0)
            {
                return;
            }
#if 0
            if (count < bestCount)
            {
                bestDigit = digit;
                bestCount = count;
                bestValid = valid;
            }
#endif
#if 1
            {
                bestDigit = digit;
                //bestCount = count;
                bestValid = valid;
                break;
            }
#endif
        }

        for (uint64_t val = 0; val < 9; val++)
        {
            if (!(bestValid & ((uint64_t)1 << val))) continue;
            ans[bestDigit] = val;
            recurse(digitsLeft - 1);
        }
        ans[bestDigit] = 9;
    }

    void solve()
    {
        uint64_t digitsLeft = digitCount;
        for (uint64_t digit = 0; digit < digitCount; digit++)
        {
            if (ans[digit] != 9) digitsLeft--;
        }
        recurse(digitsLeft);
        printf("solutionCount = %" PRIu64 "\n", solutionCount);
    }

    uint64_t digitCount = 0;
    uint64_t solutionCount = 0;
    vector<uint8_t> ans;
    vector<Clause> clauses;
};

bool isApply(Value& value)
{
    return
        value->valueType == ValueType::Binary &&
        value->binaryData.binaryOp == BinaryOp::Apply;
}

bool isLambda(Value& value)
{
    return
        value->valueType == ValueType::Lambda;
}

bool isAnd(Value& value)
{
    return
        value->valueType == ValueType::Binary &&
        value->binaryData.binaryOp == BinaryOp::And;
}
bool isOr(Value& value)
{
    return
        value->valueType == ValueType::Binary &&
        value->binaryData.binaryOp == BinaryOp::Or;
}

bool isNot(Value& value)
{
    return
        value->valueType == ValueType::Unary &&
        value->unaryData.unaryOp == UnaryOp::Not;
}

bool isVar(Value& value)
{
    return
        value->valueType == ValueType::Var;
}

bool isEq(Value& value)
{
    return
        value->valueType == ValueType::Binary &&
        value->binaryData.binaryOp == BinaryOp::Eq;
}

bool isInt(Value& value)
{
    return
        value->valueType == ValueType::Int;
}

void printValue(Value& value)
{
    string msg;

    string formatted;
    if (!formatValue(value, &formatted, &msg))
    {
        printf("%s\n", msg.c_str());
        exit(1);
    }
    printf("%s\n", formatted.c_str());
}

bool solveSat(Value& value, string* pMsg = nullptr)
{
    if (!isApply(value))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }

    Value& value2 = value->binaryData.args[0];

    if (!isApply(value2))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }

    Value& value3 = value2->binaryData.args[1];

    if (!isLambda(value3))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }
    Value& value4 = value3->lambdaData.body;

    if (!isLambda(value4))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }
    Value& value5 = value4->lambdaData.body;

    Value* pIf = &value5;
    uint64_t bitCount = 0;
    while ((*pIf)->valueType != ValueType::If)
    {
        if (!isApply(*pIf))
        {
            if (pMsg) *pMsg = "Not apply";
            return false;
        }
        pIf = &(*pIf)->binaryData.args[0];
        if (!isLambda(*pIf))
        {
            if (pMsg) *pMsg = "Not lambda";
            return false;
        }
        pIf = &(*pIf)->lambdaData.body;
        bitCount++;
    }

    //printf("bitCount = %" PRIu64 "\n", bitCount);

    auto& condValue = (*pIf)->ifData.condValue;

    //printValue(condValue);

    Sat sat;
    sat.setBitCount(bitCount);

    uint64_t clauseCount = 0;

    auto doClause = [&](Value& clause)
    {
        //printValue(clause);
        clauseCount++;

        uint64_t used = 0;
        uint64_t values = 0;

        auto doTerm = [&](Value& term)
        {
            //printValue(term);
            Value* pVar = &term;
            uint64_t value = 1;
            if (isNot(*pVar))
            {
                pVar = &(*pVar)->unaryData.arg;
                value = 0;
            }
            if (!isVar(*pVar))
            {
                printf("Not variable\n");
                exit(1);
            }

            //printf("%" PRIu64 " ", value);
            //printValue(*pVar);

            uint64_t var = (*pVar)->varData.var;
            var -= 2;
            if (var >= bitCount)
            {
                printf("var >= bitCount\n");
                exit(1);
            }

            uint64_t bit = ((uint64_t)1 << var);
            if (used & bit)
            {
                printf("Used twice\n");
                exit(1);
            }
            used |= bit;
            values |= value << var;
        };

        Value* pOr = &clause;
        while (isOr(*pOr))
        {
            doTerm((*pOr)->binaryData.args[1]);
            pOr = &(*pOr)->binaryData.args[0];
        }
        doTerm(*pOr);
        //printf("\n");
        //printf("%016" PRIx64 " %016" PRIx64 " \n", used, values);

        sat.addClause(used, values);
    };

    Value* pAnd = &condValue;
    while (isAnd(*pAnd))
    {
        doClause((*pAnd)->binaryData.args[1]);
        pAnd = &(*pAnd)->binaryData.args[0];
    }
    doClause(*pAnd);

    //printf("clauseCount = %" PRIu64 "\n", clauseCount);

    sat.solve();

    return true;
}

bool solveEq(Value& value, string* pMsg = nullptr)
{
    if (!isApply(value))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }

    Value& value2 = value->binaryData.args[0];

    if (!isApply(value2))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }

    Value& value3 = value2->binaryData.args[1];

    if (!isLambda(value3))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }
    Value& value4 = value3->lambdaData.body;

    if (!isLambda(value4))
    {
        if (pMsg) *pMsg = "Not apply";
        return false;
    }
    Value& value5 = value4->lambdaData.body;

    Value* pIf = &value5;
    uint64_t digitCount = 0;
    while ((*pIf)->valueType != ValueType::If)
    {
        if (!isApply(*pIf))
        {
            if (pMsg) *pMsg = "Not apply";
            return false;
        }
        pIf = &(*pIf)->binaryData.args[0];
        if (!isLambda(*pIf))
        {
            if (pMsg) *pMsg = "Not lambda";
            return false;
        }
        pIf = &(*pIf)->lambdaData.body;
        digitCount++;
    }

    printf("digitCount = %" PRIu64 "\n", digitCount);

    auto& condValue = (*pIf)->ifData.condValue;

    //printValue(condValue);

    Eq eq;
    eq.setDigitCount(digitCount);

    uint64_t clauseCount = 0;

    auto doClause = [&](Value& clause)
    {
        //printValue(clause);
        clauseCount++;

        if (isNot(clause))
        {
            Value& valueEq = clause->unaryData.arg;
            if (!isEq(valueEq))
            {
                printf("Not eq\n");
                exit(1);
            }

            Value& valueVar0 = valueEq->binaryData.args[0];
            Value& valueVar1 = valueEq->binaryData.args[1];
            if (!isVar(valueVar0) || !isVar(valueVar1))
            {
                printf("Not var\n");
                exit(1);
            }

            uint64_t digit0 = valueVar0->varData.var;
            uint64_t digit1 = valueVar1->varData.var;
            digit0 -= 2;
            digit1 -= 2;
            if (digit0 >= digitCount || digit1 >= digitCount)
            {
                printf("digit >= digitCount\n");
                exit(1);
            }

            eq.addClause(digit0, digit1);
        }
        else if (isEq(clause))
        {
            Value& valueVar = clause->binaryData.args[0];
            if (!isVar(valueVar))
            {
                printf("Not var\n");
                exit(1);
            }

            Value& valueVal = clause->binaryData.args[1];
            if (!isInt(valueVal))
            {
                printf("Not int\n");
                exit(1);
            }

            uint64_t digit = valueVar->varData.var;
            digit -= 2;
            if (digit >= digitCount)
            {
                printf("digit >= digitCount\n");
                exit(1);
            }

            if (!valueVal->intData.intValue.fits_ulong_p())
            {
                printf("value out of range\n");
                exit(1);
            }
            uint64_t val = valueVal->intData.intValue.get_ui();
            val--;
            if (val > 8)
            {
                printf("val > 8\n");
                exit(1);
            }

            eq.addDigit(digit, val);
        }
        else
        {
            printf("Not not or eq\n");
            exit(1);
        }
    };

    Value* pAnd = &condValue;
    while (isAnd(*pAnd))
    {
        doClause((*pAnd)->binaryData.args[1]);
        pAnd = &(*pAnd)->binaryData.args[0];
    }
    doClause(*pAnd);

    printf("clauseCount = %" PRIu64 "\n", clauseCount);

    eq.solve();

    return true;
}

void usage()
{
    printf("Usage: efficiency [<options>] <id>\n");
    printf("  <id>  Problem id\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    uint64_t id = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id> 13)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string msg;

    string problemFileName = strprintf("../problems/efficiency%" PRIu64 "", id);
    string problemData;
    if (!readFile(problemFileName, &problemData))
    {
        printf("Error reading problem file\n");
        return 1;
    }

    while (!problemData.empty() && problemData.back() == '\n')
    {
        problemData.pop_back();
    }

    printf("%s\n", problemData.c_str());

    Value value;
    if (!decodeValue(problemData, &value, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    string formatted;
    if (!formatValue(value, &formatted, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    printf("%s\n", formatted.c_str());

    if (id == 7 || id == 8)
    {
        if (!solveSat(value, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }
    }

    if (id == 9 || id == 10 || id == 11)
    {
        if (!solveEq(value, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }
    }

    return 0;
}
