#include "Common.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "SExpr.hpp"
#include "Value.hpp"
#include "Compile.hpp"
#include "Eval.hpp"
#include "Cleanup.hpp"
#include "Protocol.hpp"

using std::string;

class Problem
{
public:
    string data;
    uint64_t width = 0;
    uint64_t height = 0;
    uint64_t startX = 0;
    uint64_t startY = 0;
};

bool parseProblem(const string& problemData,
                  Problem* pProblem,
                  string* pMsg = nullptr)
{
    Problem problem;
    bool gotStart = false;
    bool done = false;
    uint64_t x = 0;
    uint64_t y = 0;
    for (char ch : problemData)
    {
        if (done)
        {
            if (ch != '\n')
            {
                if (pMsg) *pMsg = "Unexpected data at end of problem";
                return false;
            }
            continue;
        }
        if (ch == 'L')
        {
            if (gotStart)
            {
                if (pMsg) *pMsg = "Duplicate starting position";
                return false;
            }
            problem.startX = x;
            problem.startY = y;
            gotStart = true;
        }
        if (ch != '#' && ch != '.' && ch != 'L' && ch != '\n')
        {
            if (pMsg) *pMsg = "Unexpected character";
            return false;
        }
        if (problem.width == 0)
        {
            if (ch == '\n')
            {
                problem.width = x;
                x = 0;
                y++;
            }
            else
            {
                problem.data.push_back(ch);
                x++;
            }
        }
        else
        {
            if (x == problem.width)
            {
                if (ch != '\n')
                {
                    if (pMsg) *pMsg = "Newline expected";
                    return false;
                }
                x = 0;
                y++;
            }
            else
            {
                if (ch == '\n')
                {
                    if (x == 0)
                    {
                        done = true;
                    }
                    else
                    {
                        if (pMsg) *pMsg = "Newline not expected";
                        return false;
                    }
                }
                else
                {
                    problem.data.push_back(ch);
                    x++;
                }
            }
        }
    }

    if (x != 0)
    {
        if (pMsg) *pMsg = "Problem ended in middle of line";
        return false;
    }

    if (problem.width == 0)
    {
        if (pMsg) *pMsg = "Problem has zero width";
        return false;
    }

    problem.height = y;

    if (!gotStart)
    {
        if (pMsg) *pMsg = "No starting location";
        return false;
    }

    if (pProblem) *pProblem = std::move(problem);
    return true;
}

bool parseSolution(const string& solutionData,
                   string* pSolution,
                   string* pMsg = nullptr)
{
    SExpr expr;
    if (!parseSExpr(solutionData, &expr, pMsg)) return false;
    Value program;
    if (!compile(expr, &program, pMsg)) return false;
#if 0
    string strProgram;
    if (!formatValue(program, &strProgram, pMsg)) return false;
    printf("%s\n", strProgram.c_str());
#endif
    string encodedProgram;
    if (!encodeValue(program, &encodedProgram, pMsg)) return false;

    if (pSolution) *pSolution = std::move(encodedProgram);
    return true;
}

bool checkCleared(uint64_t id,
                  const Problem& problem,
                  const string& str,
                  bool *pCleared,
                  string* pMsg = nullptr)
{
    string prefix = strprintf("solve lambdaman%" PRIu64 " ", id);
    if (str.size() < prefix.size())
    {
        if (pMsg) *pMsg = "Solution too short for prefix";
        return false;
    }
    if (str.substr(0, prefix.size()) != prefix)
    {
        if (pMsg) *pMsg = "Solution does not start with prefix";
        return false;
    }

    string visited(problem.data);
    uint64_t width = problem.width;
    uint64_t height = problem.height;
    uint64_t x = problem.startX;
    uint64_t y = problem.startY;
    visited[y * width + x] = ' ';
    for (size_t i = prefix.size(); i < str.size(); i++)
    {
        char ch = str[i];
        if (ch == 'U')
        {
            if (y > 0 && visited[(y - 1) * width + x] != '#')
            {
                y--;
            }
        }
        else if (ch == 'D')
        {
            if (y + 1 < height && visited[(y + 1) * width + x] != '#')
            {
                y++;
            }
        }
        else if (ch == 'L')
        {
            if (x > 0 && visited[y * width + (x - 1)] != '#')
            {
                x--;
            }
        }
        else if (ch == 'R')
        {
            if (x + 1 < width && visited[y * width + (x + 1)] != '#')
            {
                x++;
            }
        }
        else
        {
            if (pMsg) *pMsg = "Invalid character in path";
            return false;
        }

        visited[y * width + x] = ' ';
    }

    bool cleared = true;
    for (char ch : visited)
    {
        if (ch == '.')
        {
            cleared = false;
            break;
        }
    }

#if 0
    if (!cleared)
    {
        printf("\n");
        for (uint64_t y = 0; y < height; y++)
        {
            for (uint64_t x = 0; x < width; x++)
            {
                printf("%c", visited[y * width + x]);
            }
            printf("\n");
        }
    }
#endif

    if (pCleared) *pCleared = cleared;
    return true;
}

bool calcScore(uint64_t id,
               const Problem& problem,
               const string& solution,
               uint64_t *pScore,
               string* pMsg = nullptr)
{
    Value program;
    if (!decodeValue(solution, &program, pMsg)) return false;
    Value result;
    if (!eval(program, &result, pMsg)) return false;
#if 0
    string strResult;
    if (!formatValue(result, &strResult, pMsg)) return false;
    printf("%s\n", strResult.c_str());
#endif
    if (result->valueType != ValueType::String)
    {
        if (pMsg) *pMsg = "Solution result not string";
        return false;
    }
    auto& str = result->stringData.stringValue;
    bool cleared;
    if (!checkCleared(id, problem, str, &cleared, pMsg)) return false;
    if (!cleared)
    {
        if (pMsg) *pMsg = strprintf("Grid not cleared (score would be %" PRIuZ ")", solution.size());
        return false;
    }

    if (pScore) *pScore = solution.size();
    return true;
}

void rightHandRule(const Problem& problem)
{
    uint64_t count = 0;
    for (char ch : problem.data)
    {
        if (ch == '.') count++;
    }

    string visited(problem.data);

    uint64_t width = problem.width;
    uint64_t height = problem.height;
    uint64_t x = problem.startX;
    uint64_t y = problem.startY;
    uint64_t dir = 0;

    string solution;

    while (count != 0)
    {
        switch (dir)
        {
        case 0:
        {
            if (y > 0 && visited[(y - 1) * width + x] != '#')
            {
                y--;
                dir = (dir + 1) & 3;
                solution += 'U';
            }
            else
            {
                dir = (dir + 3) & 3;
            }
            break;
        }
        case 1:
        {
            if (x + 1 < width && visited[y * width + (x + 1)] != '#')
            {
                x++;
                dir = (dir + 1) & 3;
                solution += 'R';
            }
            else
            {
                dir = (dir + 3) & 3;
            }
            break;
        }
        case 2:
        {
            if (y + 1 < height && visited[(y + 1) * width + x] != '#')
            {
                y++;
                dir = (dir + 1) & 3;
                solution += 'D';
            }
            else
            {
                dir = (dir + 3) & 3;
            }
            break;
        }
        case 3:
        {
            if (x > 0 && visited[y * width + (x - 1)] != '#')
            {
                x--;
                dir = (dir + 1) & 3;
                solution += 'L';
            }
            else
            {
                dir = (dir + 3) & 3;
            }
            break;
        }
        }
        if (visited[y * width + x] == '.')
        {
            count--;
            visited[y * width + x] = ' ';
        }

        if (x == problem.startX &&
            y == problem.startY &&
            dir == 0)
        {
            printf("Right hand rule repeating\n");
            return;
        }
    }

    //printf("\"%s\"\n", solution.c_str());
    //printf("Right hand rule: size = %" PRIuZ "\n", solution.size());

#if 0
    {
        string bits;
        for (uint64_t y = 0; y < height; y += 2)
        {
            for (uint64_t x = 1; x < width; x += 2)
            {
                char ch = problem.data[y * width + x];
                if (ch == '#') bits += '0';
                else bits += '1';
            }
        }
        std::reverse(bits.begin(), bits.end());
        printf("horiz = 0b%s\n", bits.c_str());
    }

    {
        string bits;
        for (uint64_t x = 0; x < width; x += 2)
        {
            for (uint64_t y = 1; y < height; y += 2)
            {
                char ch = problem.data[y * width + x];
                if (ch == '#') bits += '0';
                else bits += '1';
            }
        }
        std::reverse(bits.begin(), bits.end());
        printf("vert = 0b%s\n", bits.c_str());
    }

    printf("startX = %" PRIu64 "\n", (problem.startX - 1) / 2);
    printf("startY = %" PRIu64 "\n", (problem.startY - 1) / 2);
#endif
}

bool isPrime(uint64_t m)
{
    for (uint64_t d = 2; d*d <= m; d++)
    {
        if (m % d == 0) return false;
    }
    return true;
}

void solveRandom(uint64_t id,
                 const Problem& problem,
                 uint64_t maxM,
                 uint64_t stride)
{
    string msg;
    for (uint64_t m = maxM; m >= maxM / 2; m--)
    {
        if (!isPrime(m)) continue;
        printf("=== m = %" PRIu64 " ===\n", m);
        for (uint64_t g = 2; g < m; g++)
        {
            //if (g >= 94) break;
            if (g >= 8836) break;
            uint64_t order = 1;
            uint64_t power = g;
            while (power != 1)
            {
                power = (power * g) % m;
                order++;
            }
            if (order == m - 1)
            {
                printf("%4" PRIu64 ": %4" PRIu64 "\n", g, order);

                //uint64_t m_over_4 = (m + 3) / 4;

                string str;
                uint64_t n = g;
                while (n != 1)
                {
                    for (uint64_t i = 0; i < stride; i++)
                    {
                        str += "ULDR"[n % 4];
                        //str += "ULDR"[n / m_over_4];
                    }
                    n = (n * g) % m;
                }
                std::reverse(str.begin(), str.end());
                str.insert(0, strprintf("solve lambdaman%" PRIu64 " ", id));
                bool cleared;
                if (!checkCleared(id, problem, str, &cleared, &msg))
                {
                    printf("%s\n", msg.c_str());
                    exit(1);
                }
                if (cleared)
                {
                    printf("CLEARED!\n");
                    exit(0);
                }
            }
        }
    }
}

void usage()
{
    printf("Usage: lambdaman [<options>] [<id>]\n");
    printf("  <id>  Problem id (default: all)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -s    Submit solution\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    bool submit = false;
    uint64_t requestedId = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-s")
        {
            submit = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &requestedId) || requestedId < 1 || requestedId > 21)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string msg;

#if 0
    {
        //uint32_t m = 8831;
        //uint32_t m = 49999;
        //uint32_t m = 99991;
        uint32_t m = 499979;
        //uint32_t m = 830579;
        printf("=== m = %" PRIu32 " ===\n", m);
        for (uint32_t g = 2; g < m; g++)
        {
            if (g >= 94) break;
            //if (g >= 8836) break;
            uint32_t order = 1;
            uint32_t power = g;
            while (power != 1)
            {
                power = (power * g) % m;
                order++;
            }
            if (order == m - 1)
            {
                printf("%4" PRIu32 ": %4" PRIu32 "\n", g, order);
            }
        }
    }
#endif

    for (uint64_t id = 1; id <= 21; id++)
    {
        if (gotId && id != requestedId) continue;

        printf("%2" PRIu64 ": ", id);
        string problemFileName = strprintf("../problems/lambdaman%" PRIu64 "", id);
        string problemData;
        if (!readFile(problemFileName, &problemData))
        {
            printf("Error reading problem file\n");
            continue;
        }

        Problem problem;
        if (!parseProblem(problemData, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        printf("%3" PRIu64 " x %3" PRIu64 "  ", problem.width, problem.height);

#if 1
        if (id >= 11 && id <= 15)
        {
            rightHandRule(problem);
        }
#endif

#if 0
        if (id == 4)
        {
            solve4(problem);
        }
#endif
#if 0
        if (id == 11)
        {
            solveRandom(id, problem, 499990, 2);
        }
#endif
#if 0
        if (id == 16)
        {
            solveRandom(id, problem, 499990, 2);
        }
#endif
#if 0
        if (id == 17)
        {
            solveRandom(id, problem, 499990, 1);
        }
#endif
#if 1
        if (id == 18)
        {
            solveRandom(id, problem, 499990, 2);
        }
#endif
#if 0
        if (id == 21)
        {
            solveRandom(id, problem, 499990, 1);
        }
#endif

        string solutionFileName = strprintf("../solutions/lambdaman%" PRIu64 "", id);
        string solutionData;
        if (!readFile(solutionFileName, &solutionData))
        {
            printf("Error reading solution file\n");
            continue;
        }

        string solution;
        if (!parseSolution(solutionData, &solution, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }
#if 0
        printf("%s\n", solution.c_str());
#endif

        uint64_t score;
        if (!calcScore(id, problem, solution, &score, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        if (submit)
        {
            string responseStr;
            if (!Protocol::communicate(solution, &responseStr, &msg))
            {
                printf("%s\n", msg.c_str());
                continue;
            }

            Value responseValue;
            if (!decodeValue(responseStr, &responseValue, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            Value evalValue;
            if (!eval(responseValue, &evalValue, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            if (evalValue->valueType == ValueType::String)
            {
                printf("%s\n", evalValue->stringData.stringValue.c_str());
            }
            else
            {
                string evalFormatted;
                if (!formatValue(evalValue, &evalFormatted, &msg))
                {
                    printf("%s\n", msg.c_str());
                    return 1;
                }
                printf("%s\n", evalFormatted.c_str());
            }
        }
        else
        {
            printf("%" PRIu64 "\n", score);
        }
    }

#if 0
    string progStr("\"solve lambdaman1\"\n");
    SExpr progExpr;
    if (!parseSExpr(progStr, &progExpr, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    Value progValue;
    if (!compile(progExpr, &progValue, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    string valueStr;
    if (!formatValue(progValue, &valueStr, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    printf("%s\n", valueStr.c_str());
    string encodedStr;
    if (!encodeValue(progValue, &encodedStr, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    printf("%s\n", encodedStr.c_str());
    Value result;
    if (!eval(progValue, &result, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    string resultStr;
    if (!formatValue(result, &resultStr, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }
    printf("%s\n", resultStr.c_str());
#endif

    return 0;
}
