#include "Common.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "SExpr.hpp"
#include "Value.hpp"
#include "Compile.hpp"
#include "Eval.hpp"
#include "Cleanup.hpp"
#include "Protocol.hpp"
#include "ProblemCount.hpp"
#include "Problem.hpp"
#include "IsValid.hpp"
#include "Sim.hpp"
#include "DebugGraphics.hpp"

using std::string;
using std::vector;
using std::set;

bool solveRecursiveImpl(set<Coord>& coords,
                        Solution& solution,
                        Ship ship,
                        int64_t misses)
{
    if (coords.empty())
    {
        return true;
    }

    for (uint64_t pass = 0; pass < 2; pass++)
    {
        bool foundAny = false;
        for (char move = '1'; move <= '9'; move++)
        {
            Ship newShip(ship);
            newShip.step(move);
            Coord newCoord(newShip.x, newShip.y);
            auto it = coords.find(newCoord);
            if (it != coords.end())
            {
                coords.erase(it);
                solution.push_back(move);
                if (solveRecursiveImpl(coords, solution, newShip, 0)) return true;
                solution.pop_back();
                coords.insert(newCoord);
                foundAny = true;
            }
            else if (pass == 1)
            {
                solution.push_back(move);
                if (solveRecursiveImpl(coords, solution, newShip, misses + 1)) return true;
                solution.pop_back();
            }
        }
        if (foundAny || misses >= 0)
        {
            break;
        }
    }

    return false;
}

bool solveRecursive(const Problem& problem,
                    Solution* pSolution)
{
    set<Coord> coords;
    for (Coord coord : problem)
    {
        if (coord != Coord())
        {
            coords.insert(coord);
        }
    }

    string solution;
    if (!solveRecursiveImpl(coords, solution, Ship(), 0)) return false;

    if (pSolution) *pSolution = std::move(solution);
    return true;
}

bool solveSimple(const Problem& problem,
                 Solution* pSolution)
{
    set<Coord> coords;
    for (Coord coord : problem)
    {
        if (coord != Coord())
        {
            coords.insert(coord);
        }
    }

    string solution;
    Ship ship;

    while (!coords.empty())
    {
        bool found = false;
        for (char move = '1'; move <= '9'; move++)
        {
            Ship newShip(ship);
            newShip.step(move);
            //printf("(%" PRIi64 ", %" PRIi64 ")\n", newShip.x, newShip.y);
            auto it = coords.find(Coord(newShip.x, newShip.y));
            if (it != coords.end())
            {
                //printf("FOUND!\n");
                found = true;
                coords.erase(it);
                ship = newShip;
                solution.push_back(move);
                break;
            }
        }
        if (!found)
        {
            return false;
        }
    }

    if (pSolution) *pSolution = std::move(solution);
    return true;
}

class VelRange
{
public:
    int64_t low = 0;
    int64_t high = 0;
};

class PosRange
{
public:
    int64_t low = 0;
    int64_t high = 0;

    const VelRange& operator[](int64_t pos) const { return vels[pos - low]; }
    VelRange& operator[](int64_t pos) { return vels[pos - low]; }

    vector<VelRange> vels;
};

class RangePair
{
public:
    PosRange rangeX;
    PosRange rangeY;
};

void initRange(int64_t pos, int64_t vel,
               PosRange& range)
{
    range.low = pos;
    range.high = pos;
    range.vels.resize(1);
    range.vels[0].low = vel;
    range.vels[0].high = vel;
}

void initRangePair(const Ship& ship,
                   RangePair& rangePair)
{
    initRange(ship.x, ship.dx, rangePair.rangeX);
    initRange(ship.y, ship.dy, rangePair.rangeY);
}

void stepRange(const PosRange& before,
               PosRange& after)
{
    after.low = before.low + before[before.low].low - 1;
    after.high = before.high + before[before.high].high + 1;
    after.vels.resize(after.high - after.low + 1);
    for (int64_t pos = after.low; pos <= after.high; pos++)
    {
        after[pos].low = std::numeric_limits<int64_t>::max();
        after[pos].high = std::numeric_limits<int64_t>::min();
    }
    if (before.low == before.high)
    {
        int64_t pos = before.low;
        for (int64_t vel = before[pos].low; vel <= before[pos].high; vel++)
        {
            for (int64_t dv = -1; dv <= 1; dv++)
            {
                int64_t afterVel = vel + dv;
                int64_t afterPos = pos + afterVel;
                after[afterPos].low = std::min(after[afterPos].low, afterVel);
                after[afterPos].high = std::max(after[afterPos].high, afterVel);
            }
        }
    }
    else
    {
        for (int64_t pos = before.low; pos <= before.high; pos++)
        {
            int64_t lowVel = before[pos].low;
            int64_t highVel = before[pos].high;

            for (int64_t dv = -1; dv <= 1; dv++)
            {
                int64_t afterVel = lowVel + dv;
                int64_t afterPos = pos + afterVel;
                after[afterPos].low = std::min(after[afterPos].low, afterVel);
                //after[afterPos].high = std::max(after[afterPos].high, afterVel);
            }

            for (int64_t dv = -1; dv <= 1; dv++)
            {
                int64_t afterVel = highVel + dv;
                int64_t afterPos = pos + afterVel;
                //after[afterPos].low = std::min(after[afterPos].low, afterVel);
                after[afterPos].high = std::max(after[afterPos].high, afterVel);
            }
        }
    }

    for (int64_t pos = after.low; pos <= after.high; pos++)
    {
        if (after[pos].low > after[pos].high)
        {
            printf("stepRange missed some positions\n");
            exit(1);
        }
    }
};

void stepRangePair(const RangePair& before,
                   RangePair& after)
{
    stepRange(before.rangeX, after.rangeX);
    stepRange(before.rangeY, after.rangeY);
}

void restrictRange(PosRange& range, int64_t pos)
{
    VelRange velRange = range[pos];
    range.low = pos;
    range.high = pos;
    range.vels.resize(1);
    range.vels[0] = velRange;
}

void restrictRangePair(RangePair& rangePair, Coord coord)
{
    restrictRange(rangePair.rangeX, coord.x);
    restrictRange(rangePair.rangeY, coord.y);
}

class CoordIndex
{
public:
    void init(const Problem& problem)
    {
        auto localCoords = problem;
        std::sort(localCoords.begin(), localCoords.end());
        localCoords.erase(std::unique(localCoords.begin(), localCoords.end()), localCoords.end());

        int64_t minX = std::numeric_limits<int64_t>::max();
        int64_t maxX = std::numeric_limits<int64_t>::min();
        int64_t minY = std::numeric_limits<int64_t>::max();
        int64_t maxY = std::numeric_limits<int64_t>::min();
        for (Coord coord : localCoords)
        {
            if (coord == Coord()) continue;
            minX = std::min(minX, coord.x);
            maxX = std::max(maxX, coord.x);
            minY = std::min(minY, coord.y);
            maxY = std::max(maxY, coord.y);
        }

        bucketLowX = minX;
        bucketLowY = minY;
        bucketSizeX = (uint64_t)(maxX - minX + 1 + 31) / 32;
        bucketSizeY = (uint64_t)(maxY - minY + 1 + 31) / 32;

        for (Coord coord : localCoords)
        {
            if (coord == Coord()) continue;
            uint64_t bucketX = (uint64_t)(coord.x - bucketLowX) / bucketSizeX;
            uint64_t bucketY = (uint64_t)(coord.y - bucketLowY) / bucketSizeY;
            data[bucketY][bucketX].push_back(coord);
        }
    }

    void find(const RangePair& rangePair,
              vector<Coord>& coords)
    {
        coords.clear();

        int64_t lowX = rangePair.rangeX.low;
        int64_t highX = rangePair.rangeX.high;
        int64_t lowY = rangePair.rangeY.low;
        int64_t highY = rangePair.rangeY.high;
        if (highX < bucketLowX) return;
        if (highY < bucketLowY) return;
        if (lowX < bucketLowX) lowX = bucketLowX;
        if (lowY < bucketLowY) lowY = bucketLowY;

        uint64_t lowBucketX = (uint64_t)(lowX - bucketLowX) / bucketSizeX;
        uint64_t highBucketX = (uint64_t)(highX - bucketLowX) / bucketSizeX;
        uint64_t lowBucketY = (uint64_t)(lowY - bucketLowY) / bucketSizeY;
        uint64_t highBucketY = (uint64_t)(highY - bucketLowY) / bucketSizeY;

        if (lowBucketX > 31) return;
        if (lowBucketX > 31) return;
        if (highBucketX > 31) highBucketX = 31;
        if (highBucketY > 31) highBucketY = 31;

        for (uint64_t bucketY = lowBucketY; bucketY <= highBucketY; bucketY++)
        {
            for (uint64_t bucketX = lowBucketX; bucketX <= highBucketX; bucketX++)
            {
                for (Coord coord : data[bucketY][bucketX])
                {
                    if (coord.x >= rangePair.rangeX.low &&
                        coord.x <= rangePair.rangeX.high &&
                        coord.y >= rangePair.rangeY.low &&
                        coord.y <= rangePair.rangeY.high)
                    {
                        coords.push_back(coord);
                    }
                }
            }
        }
    }

public:
    vector<Coord> data[32][32];
    int64_t bucketLowX = 0;
    int64_t bucketLowY = 0;
    uint64_t bucketSizeX = 0;
    uint64_t bucketSizeY = 0;
};

void reconstructOne(const PosRange& before,
                    const PosRange& after,
                    int64_t posAfter,
                    int64_t velAfter,
                    int64_t& posBefore,
                    int64_t& velBefore,
                    int64_t& move)
{
    posBefore = posAfter - velAfter;
    if (posBefore < before.low ||
        posBefore > before.high)
    {
        printf("Reconstruct position error!\n");
        exit(1);
    }
    auto& velRange = before[posBefore];
    if (velAfter >= velRange.low &&
        velAfter <= velRange.high)
    {
        velBefore = velAfter;
        move = 0;
    }
    else if (velAfter == velRange.low - 1)
    {
        velBefore = velAfter + 1;
        move = -1;
    }
    else if (velAfter == velRange.high + 1)
    {
        velBefore = velAfter - 1;
        move = 1;
    }
    else
    {
        printf("Reconstruct velocity error!\n");
        exit(1);
    }
}

void reconstruct(const vector<RangePair>& rangePath,
                 string& solution)
{
    size_t initialPos = solution.size();

    uint64_t size = rangePath.size() - 1;

    Ship ship;
    ship.x = rangePath[size].rangeX.low;
    ship.dx = rangePath[size].rangeX[ship.x].low;
    ship.y = rangePath[size].rangeY.low;
    ship.dy = rangePath[size].rangeY[ship.y].low;

    for (uint64_t idx = size; idx > 0; )
    {
        idx--;
        auto& pairBefore = rangePath[idx];
        auto& pairAfter = rangePath[idx + 1];
        //printf("after (%" PRIi64 " .. %" PRIi64 ", %" PRIi64 " .. %" PRIi64 ")\n", pairAfter.rangeX.low, pairAfter.rangeX.high, pairAfter.rangeY.low, pairAfter.rangeY.high);
        //printf("before (%" PRIi64 " .. %" PRIi64 ", %" PRIi64 " .. %" PRIi64 ")\n", pairBefore.rangeX.low, pairBefore.rangeX.high, pairBefore.rangeY.low, pairBefore.rangeY.high);
        //printf("ship:    x = %" PRIi64 ", dx = %" PRIi64 ", y = %" PRIi64 ", dy = %" PRIi64 "\n", ship.x, ship.dx, ship.y, ship.dy);
        Ship newShip;
        int64_t moveX = 0;
        int64_t moveY = 0;
        reconstructOne(pairBefore.rangeX, pairAfter.rangeX, ship.x, ship.dx, newShip.x, newShip.dx, moveX);
        reconstructOne(pairBefore.rangeY, pairAfter.rangeY, ship.y, ship.dy, newShip.y, newShip.dy, moveY);
        //printf("newShip: x = %" PRIi64 ", dx = %" PRIi64 ", y = %" PRIi64 ", dy = %" PRIi64 "\n", newShip.x, newShip.dx, newShip.y, newShip.dy);

        char move = '1' + ((moveY + 1) * 3 + (moveX + 1));
        solution.push_back(move);
        ship = newShip;
    }

    std::reverse(solution.begin() + initialPos, solution.end());
}

bool solveGreedy(const Problem& problem,
                 Solution* pSolution)
{
    vector<Coord> coordsLeft;
    coordsLeft.reserve(problem.size());
    for (Coord coord : problem)
    {
        if (coord != Coord())
        {
            coordsLeft.push_back(coord);
        }
    }
    std::sort(coordsLeft.begin(), coordsLeft.end());
    coordsLeft.erase(std::unique(coordsLeft.begin(), coordsLeft.end()), coordsLeft.end());

    Coord nextCoord;
    uint64_t nextCoordIdx = (uint64_t)-1;

    //CoordIndex index;
    //index.init(problem);

    vector<RangePair> rangePath;

    rangePath.emplace_back();
    initRangePair(Ship(), rangePath.back());

    vector<Coord> coords;

    uint64_t size = 0;

    vector<Coord> fakeSolution;
    fakeSolution.emplace_back();

    bool useDebugGraphics = false;

    DebugGraphics debugGraphics;
    string text;
    if (useDebugGraphics)
    {
        debugGraphics.create();
        debugGraphics.update(problem, fakeSolution, text);
    }

    while (!coordsLeft.empty())
    {
        if (nextCoordIdx == (uint64_t)-1)
        {
            int64_t bestDist = std::numeric_limits<int64_t>::max();
            uint64_t bestIdx = 0;

            for (uint64_t idx = 0; idx < coordsLeft.size(); idx++)
            {
                Coord coord = coordsLeft[idx];
                int64_t dist =
                    std::abs(coord.x - nextCoord.x) +
                    std::abs(coord.y - nextCoord.y);
                if (dist < bestDist)
                {
                    bestDist = dist;
                    bestIdx = idx;
                }
            }

            nextCoordIdx = bestIdx;
            nextCoord = coordsLeft[bestIdx];
        }

        size++;
        rangePath.emplace_back();
        auto& beforeRangePair = rangePath[size - 1];
        auto& afterRangePair = rangePath[size];
        //printf("BEFORE (%" PRIi64 " .. %" PRIi64 ", %" PRIi64 " .. %" PRIi64 ")\n", beforeRangePair.rangeX.low, beforeRangePair.rangeX.high, beforeRangePair.rangeY.low, beforeRangePair.rangeY.high);
        stepRangePair(beforeRangePair, afterRangePair);
        //printf("AFTER (%" PRIi64 " .. %" PRIi64 ", %" PRIi64 " .. %" PRIi64 ")\n", afterRangePair.rangeX.low, afterRangePair.rangeX.high, afterRangePair.rangeY.low, afterRangePair.rangeY.high);

#if 1
        if (nextCoord.x >= afterRangePair.rangeX.low &&
            nextCoord.x <= afterRangePair.rangeX.high &&
            nextCoord.y >= afterRangePair.rangeY.low &&
            nextCoord.y <= afterRangePair.rangeY.high)
        {
            coordsLeft[nextCoordIdx] = coordsLeft.back();
            coordsLeft.pop_back();
            nextCoordIdx = (uint64_t)-1;

            restrictRangePair(afterRangePair, nextCoord);

            fakeSolution.push_back(nextCoord);
            if (useDebugGraphics)
            {
                debugGraphics.update(fakeSolution, text);
            }

            //printf("."); fflush(stdout);
        }
#endif

#if 0
        index.find(afterRangePair, coords);
        //printf("index found %" PRIuZ "\n", coords.size());

        for (Coord coord : coords)
        {
            auto it = coordsLeft.find(coord);
            if (it != coordsLeft.end())
            {
                coordsLeft.erase(it);
                restrictRangePair(afterRangePair, coord);

                fakeSolution.push_back(coord);
                if (useDebugGraphics)
                {
                    debugGraphics.update(fakeSolution, text);
                }

                printf("."); fflush(stdout);
                break;
            }
        }
#endif

#if 0
        for (Coord coord : index.data)
        {
            auto it = coordsLeft.find(coord);
            if (it != coordsLeft.end())
            {
                coordsLeft.erase(it);
                restrictRangePair(afterRangePair, coord);
                break;
            }
        }
#endif

#if 0
        if (useDebugGraphics)
        {
            if (debugGraphics.destroyed())
            {
                break;
            }
        }
#endif
    }

    if (useDebugGraphics)
    {
        debugGraphics.destroy();
    }

    string solution;
    reconstruct(rangePath, solution);

    if (pSolution) *pSolution = std::move(solution);
    return true;
}

void usage()
{
    printf("Usage: spaceship [<options>] [<id>]\n");
    printf("  <id>  Problem id (default: all)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -s    Submit solution\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    bool submit = false;
    uint64_t requestedId = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-s")
        {
            submit = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &requestedId) || requestedId < 1 || requestedId > problemCount)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    Protocol::init();
    Cleanup cleanupProtocol([](){ Protocol::cleanup(); });

    string msg;

    for (uint64_t id = 1; id <= problemCount; id++)
    {
        if (gotId && id != requestedId) continue;

        printf("%2" PRIu64 ": ", id);
        string problemFileName = strprintf("../problems/spaceship%" PRIu64 "", id);
        string problemData;
        if (!readFile(problemFileName, &problemData))
        {
            printf("Error reading problem file\n");
            continue;
        }

        vector<Coord> problem;
        if (!parseProblem(problemData, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        printf("%5" PRIuZ "  ", problem.size());

        int64_t minX = 0;
        int64_t maxX = 0;
        int64_t minY = 0;
        int64_t maxY = 0;
        for (Coord coord : problem)
        {
            minX = std::min(minX, coord.x);
            maxX = std::max(maxX, coord.x);
            minY = std::min(minY, coord.y);
            maxY = std::max(maxY, coord.y);
        }

        printf("%7" PRIi64" %7" PRIi64 "  ", maxX - minX + 1, maxY - minY + 1);

        fflush(stdout);

#if 1
        {
            string solution;
            //if (solveSimple(problem, &solution))
            //if (solveRecursive(problem, &solution))
            if (solveGreedy(problem, &solution))
            {
                if (!isValid(problem, solution, &msg))
                {
                    printf("%s\n", msg.c_str());
                    continue;
                }

                uint64_t score = solution.size();
                printf("%" PRIu64 "\n", score);

#if 0
                if (!writeSolution(id, solution))
                {
                    printf("Error writing solution\n");
                    return 1;
                }
#endif
                //printf("%s\n", solution.c_str());
            }
            else
            {
                printf("not solved\n");
            }
            continue;
        }
#endif

        string solutionFileName = strprintf("../solutions/spaceship%" PRIu64 "", id);
        string solution;
        if (!readFile(solutionFileName, &solution))
        {
            printf("Error reading solution file\n");
            continue;
        }

        if (!isValid(problem, solution, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        uint64_t score = solution.size();

        if (submit)
        {
            Value value;
            value.init();
            value->setValueType(ValueType::String);
            value->stringData.stringValue = strprintf("solve spaceship%" PRIu64 " ", id);
            value->stringData.stringValue += solution;

            string requestStr;
            if (!encodeValue(value, &requestStr, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            string responseStr;
            if (!Protocol::communicate(requestStr, &responseStr, &msg))
            {
                printf("%s\n", msg.c_str());
                continue;
            }

            Value responseValue;
            if (!decodeValue(responseStr, &responseValue, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            Value evalValue;
            if (!eval(responseValue, &evalValue, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }

            if (evalValue->valueType == ValueType::String)
            {
                printf("%s\n", evalValue->stringData.stringValue.c_str());
            }
            else
            {
                string evalFormatted;
                if (!formatValue(evalValue, &evalFormatted, &msg))
                {
                    printf("%s\n", msg.c_str());
                    return 1;
                }
                printf("%s\n", evalFormatted.c_str());
            }
        }
        else
        {
            printf("%" PRIu64 "\n", score);
        }
    }

    return 0;
}
